package com.objsys.asn1c;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.action.IStatusLineManager;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IViewReference;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

/**
 * The activator class controls the plug-in life cycle
 */
public class Activator extends AbstractUIPlugin {

	// The plug-in ID
	public static final String PLUGIN_ID = "com.objsys.asn1c"; //$NON-NLS-1$

	// The shared instance
	private static Activator plugin;

	/**
	 * The constructor
	 */
	public Activator() {
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
		plugin = this;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext context) throws Exception {
		plugin = null;
		super.stop(context);
	}

	/**
	 * Returns the shared instance
	 *
	 * @return the shared instance
	 */
	public static Activator getDefault() {
		return plugin;
	}

	/**
	 * Returns an image descriptor for the image file at the given
	 * plug-in relative path
	 *
	 * @param path the path
	 * @return the image descriptor
	 */
	public static ImageDescriptor getImageDescriptor(String path) {
		return imageDescriptorFromPlugin(PLUGIN_ID, path);
	}

	public static void log(IStatus status) {
    getDefault().getLog().log(status);
  }

  public static void log(CoreException ex) {
    final IStatus status = ex.getStatus();
    log(new Status(status.getSeverity(), PLUGIN_ID, status.getCode(), status.getMessage(), ex));
  }

  public static void log(String message) {
    log(message, 1, null);
  }

  public static void log(String message, Throwable e) {
    log(message, 1, e);
  }

  public static void log(String message, int code, Throwable e) {
    log(new Status(IStatus.ERROR, PLUGIN_ID, code, message, e));
  }

  public static void deleteMarkers(IResource resource, String type) {
    if (resource == null)
      return;
    try {
      resource.deleteMarkers(type, true, IResource.DEPTH_INFINITE);
    } catch (CoreException e) {
      System.err.println("unable to delete error markers for resource " + resource.getFullPath()); //$NON-NLS-1$
    }
  }

  public static void createMarker(
      IResource resource,
      String kind,
      int severity,
      String message,
      int start,
      int end,
      int line) {

    IMarker marker;
    try {
      marker = resource.createMarker(kind);
      marker.setAttributes(
          new String[] { IMarker.SEVERITY, IMarker.MESSAGE, IMarker.LINE_NUMBER },
          new Object[] { new Integer(severity), message, new Integer(line) });
      if (start >= 0) {
        marker.setAttribute(IMarker.CHAR_START, start);
      }
      if (end >= 0) {
        marker.setAttribute(IMarker.CHAR_START, end);
      }
    } catch (CoreException ex) {
      log( "Unable to create marker"); //$NON-NLS-1$
      return;
    }
  }


  /**
   * @param message the message to be displayed.
   * @param kind one of ({@link Status#ERROR} or {@link Status#OK})
   */
  public static void showStatusLineMessage(final String message, final int kind) {
    runInGuiThread(new Runnable() {

      public void run() {
        // Retrieves status line manager instance and sets the message into the instance.
        IWorkbenchWindow activeWindow = getDefault().getWorkbench().getActiveWorkbenchWindow();
        if (activeWindow != null) {
          IWorkbenchPage activePage = activeWindow.getActivePage();
          if (activePage != null) {
            showStatusMessage(activePage);
          }
        }
      }

      /**
       * Show the status message.
       *
       * @param activePage The active workbench.
       */
      private void showStatusMessage(IWorkbenchPage activePage) {
        IViewReference[] viewReferences = activePage.getViewReferences();
        for (int i = 0; i < viewReferences.length; i++) {
          IViewPart viewPart = viewReferences[i].getView(true);
          if (viewPart != null) {
            IActionBars viewActionbars = viewPart.getViewSite().getActionBars();
            IStatusLineManager viewStatusManager = viewActionbars.getStatusLineManager();
            // Set status line associated with view part such as Package Explore, and etc.
            switch (kind) {
            case Status.ERROR:
              viewStatusManager.setMessage(null);
              viewStatusManager.setErrorMessage(message);
              break;
            default:
              viewStatusManager.setErrorMessage(null);
              viewStatusManager.setMessage(message);
              break;
            }
          }
        }
        IEditorPart editorPart = activePage.getActiveEditor();
        if (editorPart != null) {
          IActionBars partActionBars = editorPart.getEditorSite().getActionBars();
          IStatusLineManager partStatusManager = partActionBars.getStatusLineManager();
          // Set status line associated with editor part such as text editor.
          switch (kind) {
          case Status.ERROR:
            partStatusManager.setMessage(null);
            partStatusManager.setErrorMessage(message);
            break;
          default:
            partStatusManager.setErrorMessage(null);
            partStatusManager.setMessage(message);
            break;
          }
        }
      }
    }, false);
  }

  private static void runInGuiThread(Runnable runnable, final boolean sync) {
    if (runnable == null) {
      return;
    }
    Display display = PlatformUI.getWorkbench().getDisplay();

    if (display != null && !display.isDisposed()) {
      try {
        if (sync) {
          display.syncExec(runnable);
        } else {
          display.asyncExec(runnable);
        }
      }
      catch (final SWTException e) {
        log("error ", e); //$NON-NLS-1$
      }
    }
  }
}
