package com.objsys.asn1c.popup.actions;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowActionDelegate;

/**
 * This is a base class for actions which operate on a project.  All actions
 * derived from this class will operate on the same project.
 */
public abstract class AbstractProjectActionDelegate 
   implements IWorkbenchWindowActionDelegate, IObjectActionDelegate
{
   //The project on which the next action will be run.
   protected static IProject project = null;
   
   //The last thing that we were notified was selected.
   protected ISelection selection;
       
   //For a regular menu or toolbar button, the window that the action was
   //initialized on.
   protected IWorkbenchWindow window = null;
  public AbstractProjectActionDelegate() {
    super();
  }

  protected abstract void process(IProject res);

  public void run(IAction action)
  {   
     if ( AbstractProjectActionDelegate.project == null ) {
        MessageDialog.openInformation(window.getShell(), 
              "Inappropriate selection", 
              "No project selected and no prior project has been acted on.\n" +
              "Select a project.");
     }
     else {
        process(AbstractProjectActionDelegate.project);
     }
  }
  
  public void init(IWorkbenchWindow window) {
     this.window = window;
  }
  
  public void setActivePart(IAction action, IWorkbenchPart targetPart) 
  {} //no action required
  

  public void selectionChanged(IAction action, ISelection selection)
  {
     this.selection = selection;
     
     AbstractProjectActionDelegate.project = getProjectFromSelection();
  }
  

  /**
   * This tries to return a project based on the selected item.  
   * 
   * The algorithm is as follows:
   *  if the current select has a corresponding project, return it
   *  else if some prior selection had an associated project, and that project
   *     is still open, return it
   *  else return null
   * 
   * Note that this returns only a single project: we are assuming that
   * selections are limited to a single thing so that there can be at most
   * one corresponding project.
   */
  private IProject getProjectFromSelection() {
     
     IProject project = null;     //project to return
     
     if ( selection == null ) return null;
     
     if (selection instanceof IStructuredSelection) {

        Object[] selections = ((IStructuredSelection) selection).toArray();

        for (Object selection : selections) {
           if ( selection instanceof IResource ) {
              project = ((IResource)selection).getProject();
              if ( project != null ) break;
           }
           else if ( selection instanceof IAdaptable ) {
             IAdaptable selectionAsAdaptable = (IAdaptable) selection;
             IResource selectedResource = null;
             
             //If the IAdapatble is some IResource, then we can get the project
             selectedResource = 
                (IResource) selectionAsAdaptable.getAdapter(IResource.class);
             if ( selectedResource != null ) {
                project = selectedResource.getProject();
                if ( project != null ) break;
             }
          } 
          //else: something we don't know about
        }
     }
     //else: not possible to have a non-structured selection?
     
     if ( project != null )
        AbstractProjectActionDelegate.project = project;
     
     return AbstractProjectActionDelegate.project;    //possibly null
  }
 
   public void dispose() {
      window = null;
   }  
}