package com.objsys.asn1c.validator;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Path;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;

import com.objsys.asn1c.Activator;


public class LineProcessor {

  private static final String ASN1_C = Messages.LineProcessor_console_title;

  private static final String INFO_PREFIX = "info"; //$NON-NLS-1$
  private static final String WARNING_PREFIX = "warning"; //$NON-NLS-1$
  private static final String ERROR_PREFIX = "error"; //$NON-NLS-1$

  private final IResource resource;
  private final Map<Pattern, Integer> patterns = createPatterns();
  private MessageConsole console;
  private MessageConsoleStream messageStream;

  public LineProcessor(IResource resource) {
    this.resource = resource;
    console = findConsole(ASN1_C);
    console.activate();
    messageStream = console.newMessageStream();
  }

  /**
   * @return the sorted and compiled patterns from "pattern.properties"
   */
  private static Map<Pattern, Integer> createPatterns() {
    Map<Pattern, Integer> patterns = new Hashtable<Pattern, Integer>();
    Properties properties = new Properties();
    try {
      InputStream inputStream = LineProcessor.class.getClassLoader().getResourceAsStream("/pattern.properties"); //$NON-NLS-1$
      if (inputStream != null) {
        properties.load(inputStream);
      }
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    for ( Map.Entry<Object,Object> entry : properties.entrySet()) {
      String key = (String) entry.getKey();
      String value = (String) entry.getValue();
      if (key.startsWith(WARNING_PREFIX)) {
        patterns.put(Pattern.compile(value), IMarker.SEVERITY_WARNING);
      } else if (key.startsWith(ERROR_PREFIX)) {
        patterns.put(Pattern.compile(value), IMarker.SEVERITY_ERROR);
      } else if (key.startsWith(INFO_PREFIX)) {
        patterns.put(Pattern.compile(value), IMarker.SEVERITY_INFO);
      }
    }
    return patterns;
  }

  public void process(String line) {
    if (line == null) {
      return;
    }
    printlnConsole(line);
    if ((line = line.trim()).length() == 0) {
      return;
    }
    for (Map.Entry<Pattern,Integer> entry : patterns.entrySet()) {
      Matcher matcher = entry.getKey().matcher(line);
      if (!matcher.matches()) {
        continue;
      }
      int groups = matcher.groupCount();

      String fileStr = null;
      String lineStr = null;
      String message;

      if (groups > 2) {
        // file, line and message
        fileStr = matcher.group(1);
        if (fileStr.trim().length() == 0) {
          fileStr = null;
        }
        lineStr = matcher.group(2);
        message = matcher.group(3);
      } else if (groups > 0) {
        // no line or file info
        message = matcher.group(1);
      } else {
        // entire string
        message = matcher.group(0);
      }
      int lineNo;
      try {
        lineNo = Integer.parseInt(lineStr);
      } catch (NumberFormatException e) {
        lineNo = 1;
      }
      IResource file = null;
      if (fileStr != null) {
        try {
          File absLineFile = new File(fileStr);
          if (absLineFile.exists() && absLineFile.isAbsolute()) {
            IFile[] files = ResourcesPlugin.getWorkspace().getRoot().findFilesForLocationURI(absLineFile.toURI());
            if (files.length > 0) {
              file = files[0];
            }
          }
        } catch (Exception e) {
          file = null;
        }
        if (file == null) {
          try {
            file = resource.getProject().findMember(new Path(fileStr));
          } catch (Exception e) {
            file = resource;
          }
        }
      }
      if (file != null && file.exists()) {
        Activator.createMarker(file, IMarker.PROBLEM, entry.getValue(), message, -1, -1, lineNo);
      }
    }

  }

  public void printlnConsole(String text) {
    if (messageStream.isClosed()) {
      return;
    }
    messageStream.println(text);
    try {
      messageStream.flush();
    } catch (IOException e1) {
      /* nothing */
    }
  }

  public void close() {
    try {
      messageStream.flush();
    } catch (IOException e1) {
      /* nothing */
    }
    try {
      messageStream.close();
    } catch (IOException e1) {
      /* nothing */
    }
  }

  private MessageConsole findConsole(String name) {
    ConsolePlugin plugin = ConsolePlugin.getDefault();
    IConsoleManager conMan = plugin.getConsoleManager();
    IConsole[] existing = conMan.getConsoles();
    for (int i = 0; i < existing.length; i++)
      if (name.equals(existing[i].getName()))
        return (MessageConsole) existing[i];
    // no console found, so create a new one
    MessageConsole myConsole = new MessageConsole(name, null);
    conMan.addConsoles(new IConsole[] { myConsole });
    return myConsole;
  }

  public void println(String string) {
    messageStream.println(string);
  }
}
