package com.objsys.asn1c.validator;
import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.osgi.util.NLS;
import com.objsys.asn1c.Activator;
import com.objsys.asn1c.validator.preferences.ASN1CPreferenceConstants;
import com.objsys.asn1c.validator.preferences.ASN1CPreferenceConstants.Language;
import com.objsys.asn1c.validator.project.ACProject;


public class ValidationUtils {
  /**
   * validate resource and report warnings/errors
   *
   * @param resource
   */
  public static void validate(IResource resource) {
    process(resource, true);
  }
  /**
   * process resource generate files and report warnings/errors
   *
   * @param resource
   */
  public static void generate(IResource resource) {
    process(resource, false);
  }
  /**
   * process resource and report warnings/errors
   *
   * @param resource
   */
  private static synchronized void process(final IResource resource, final boolean validation) {
    new Job(NLS.bind(Messages.ValidationUtils_processing, resource.getName())){
      protected IStatus run(IProgressMonitor monitor) {
        final IProject project = resource.getProject();
        File projectLocation = project.getLocation().toFile().getAbsoluteFile();
        String resourceFullPath = resource.getFullPath().toPortableString();
        Activator.showStatusLineMessage(null, Status.ERROR);
        LineProcessor lineProcessor = new LineProcessor(resource);
        lineProcessor.printlnConsole(NLS.bind(Messages.ValidationUtils_starting, new Date()));
        String[] files = null;
        if (resource.getType() == IResource.PROJECT) {
        } else {
          IPath path = resource.getFullPath();
          if (path == null) {
            lineProcessor.printlnConsole(NLS.bind(Messages.ValidationUtils_invalid_path, resourceFullPath));
            Activator.showStatusLineMessage(Messages.ValidationUtils_validator_error, Status.WARNING);
            return new Status(IStatus.WARNING, Activator.PLUGIN_ID, IStatus.WARNING, Messages.ValidationUtils_validator_error, null);
          }
          if (!resource.exists()) {
            lineProcessor.printlnConsole(NLS.bind(Messages.ValidationUtils_file_not_found, resourceFullPath));
            Activator.showStatusLineMessage(Messages.ValidationUtils_file_not_found, Status.WARNING);
            return new Status(IStatus.WARNING, Activator.PLUGIN_ID, IStatus.WARNING, Messages.ValidationUtils_file_not_found, null);
          }
          files = new String[]{resource.getFullPath().toPortableString()};
        }
        ACProject acPrj = new ACProject(project);
        try {
          acPrj.loadPreferenceStore();
          deleteMarkersOnInputFiles(acPrj);
          
          Collection<String> parameters = getParameters(
              acPrj, files, validation);
          if (parameters == null) {
            lineProcessor.println(Messages.ValidationUtils_validator_error);
            Activator.showStatusLineMessage(Messages.ValidationUtils_validator_error, Status.WARNING);
            return new Status(IStatus.WARNING, Activator.PLUGIN_ID, IStatus.WARNING, Messages.ValidationUtils_validator_error, null);
          }
          lineProcessor.println("Current directory: "+projectLocation); //$NON-NLS-1$
          // print exec line
          StringBuilder sb = new StringBuilder("#/> "); //$NON-NLS-1$
          for (String param : parameters) {
            sb.append(param).append(' ');
          }
          lineProcessor.println(sb.toString());
          // exec process
          String[] cmdLine = parameters.toArray(new String[0]);
          Process process = Runtime.getRuntime().exec(cmdLine, null, projectLocation);
          StreamHandler outHandler = new StreamHandler(process.getInputStream(), lineProcessor);
          StreamHandler errHandler = new StreamHandler(process.getErrorStream(), lineProcessor);
          Thread outThread = new Thread(outHandler);
          Thread errThread = new Thread(errHandler);
          outThread.start();
          errThread.start();
          int exitCode = process.waitFor();
          outThread.join();
          errThread.join();
          final String message;
          if (exitCode == 0) {
            message = NLS.bind(Messages.ValidationUtils_successful, new Date());
            lineProcessor.println(message);
          } else {
            message = NLS.bind(Messages.ValidationUtils_exit_code, exitCode, new Date());
            lineProcessor.println(message);
          }
          lineProcessor.println(""); //$NON-NLS-1$
          lineProcessor.close();
          new Job(NLS.bind(Messages.ValidationUtils_building_project, project.getName())){
            @Override
            protected IStatus run(IProgressMonitor monitor) {
              try {
                project.refreshLocal(IResource.DEPTH_INFINITE, monitor);
                project.build(IncrementalProjectBuilder.INCREMENTAL_BUILD, monitor);
              } catch (Exception e) {
                return new Status(IStatus.ERROR, Activator.PLUGIN_ID, IStatus.ERROR, e.getMessage(), e);
              }
              return new Status(IStatus.OK, Activator.PLUGIN_ID, IStatus.OK, message, null);
            }
          }.schedule();
          return new Status(IStatus.OK, Activator.PLUGIN_ID, IStatus.OK, message, null);
        } catch (Exception e) {
          lineProcessor.printlnConsole(e.getMessage());
          lineProcessor.close();
          return new Status(IStatus.OK, Activator.PLUGIN_ID, IStatus.OK, e.getMessage(), e);
//          Activator.showStatusLineMessage(e.getMessage(), Status.ERROR);
        }
      }
    }.schedule();
  }
  /**
   * 
   * @param acproject
   * @param filenames
   * @param validate Pass true if the command line for just validating is desired.
   * @return
   * @throws IOException
   */
  private static Collection<String> getParameters(ACProject acproject, String[] filenames, boolean validate) throws IOException {
    return getParameters(acproject, null, filenames, validate, false);
  }
  /**
   * This deletes all problem markers (and subtypes of problem) on the input 
   * files from acproject.
   * 
   * Deleting all markers in the project is probably deleting too much, as it
   * may remove markers for problems on files that will neither be processed
   * by asn1c nor changed, so that markers for real problems would be lost,
   * since the generate action doesn't necessarily trigger Java compilation on 
   * all .java files in the project (and the validate action never does).
   * 
   * @param acproject
   */
  private static void deleteMarkersOnInputFiles(ACProject acproject)
  {
     String[] filenames;
     
     filenames = acproject.getCollection(ASN1CPreferenceConstants.INPUT_FILES, 
           ACProject.PATH_SEPARATOR).toArray(new String[0]);
     
     for (int i = 0; i < filenames.length; i++) {
        String filename = filenames[i];

        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IResource resource = root.getFile(new Path(filename));
        try {
           resource.deleteMarkers(IMarker.PROBLEM, true, 
                 IResource.DEPTH_INFINITE);
        }
        catch (CoreException e) {
           System.err.println("unable to delete error markers for resource " + //$NON-NLS-1$
                 resource.getFullPath()); 
        }
     }
  }
  
  /**
   * 
   * @param acproject If not null, the preference store and list of file names
   *    are taken from this (ignoring the passed parameters store and filenames)
   * @param store
   * @param filenames
   * @param validate
   * @param ignoreErrors
   * @return
   * @throws IOException
   */
  private static Collection<String> getParameters(ACProject acproject, IPreferenceStore store, String[] filenames, boolean validate, boolean ignoreErrors) throws IOException {
    Collection<String> result = new LinkedList<String>();
    if (acproject != null) {
      store = acproject.getPreferenceStore();
      String command = store.getString(ASN1CPreferenceConstants.COMPILER_PATH);
      if (command == null || command.trim().length() == 0) {
        if (!ignoreErrors)
        throw new IOException(Messages.ValidationUtils_no_compiler);
      }
      if (!new File(command).exists()) {
        if (!ignoreErrors)
        throw new IOException(NLS.bind(Messages.ValidationUtils_compiler_invalid, command));
      }
      result.add(command);
      if (filenames == null) {
        filenames = acproject.getCollection(ASN1CPreferenceConstants.INPUT_FILES, ACProject.PATH_SEPARATOR).toArray(new String[0]);
      }
      if (filenames == null || filenames.length == 0) {
        if (!ignoreErrors)
        throw new IOException(Messages.ValidationUtils_no_input_files);
      }
      LinkedList<String> fsFiles = new LinkedList<String>();
      for (String filename : filenames) {
        String relativePath = filename;
        try {
          relativePath = getRelativeFSPathFromResource(acproject, filename, true, Messages.ValidationUtils_input_resource_not_found);
        } catch (RuntimeException e) {
          if (!ignoreErrors)
            throw e;
          relativePath = filename;
        } catch (IOException e) {
          if (!ignoreErrors)
            throw e;
          relativePath = filename;
        }
        fsFiles.add(relativePath);
      }
      result.addAll(fsFiles);
    }
//    // options
//
//    final Collection<String> keys;
//
//    if (validate) {
//
//      keys = ASN1CPreferenceConstants.getValidationPreferenceKeys();
//
//    } else {
//
//      String language = store.getString(ASN1CPreferenceConstants.LANGUAGE);
//
//      boolean xsd = store.getBoolean(ASN1CPreferenceConstants.XSD);
//
//      boolean symbian = store.getBoolean(ASN1CPreferenceConstants.SYMBIAN);
//
//      //TODO: fix the language to syntax checking if appropriate
//      keys = ASN1CPreferenceConstants.getPreferenceKeys(language, xsd, symbian);
//
//    }
//
//
//
//    // options
//
//    keys.remove(ASN1CPreferenceConstants.COMPILER_PATH);
//
//    keys.remove(ASN1CPreferenceConstants.INPUT_FILES);
//
//    keys.remove(ASN1CPreferenceConstants.IMPORT);
//
//    keys.remove(ASN1CPreferenceConstants.OUTPUT_DIR);
    // output dir
    String outputDir = store.getString(ASN1CPreferenceConstants.OUTPUT_DIR);
    if (!validate && outputDir != null && outputDir.trim().length() > 0) {
      result.add("-o");
      String relativePath = outputDir;
      try {
        relativePath = getRelativeFSPathFromResource(acproject, outputDir, false, Messages.ValidationUtils_output_folder_not_found);
      } catch (RuntimeException e) {
        if (!ignoreErrors)
          throw e;
        relativePath = outputDir;
      } catch (IOException e) {
        if (!ignoreErrors)
          throw e;
        relativePath = outputDir;
      }
      result.add(relativePath);
    }
    // process import dirs (workspace resource vs. file system)
    Collection<String> collection = ACProject.getCollection(store, ASN1CPreferenceConstants.IMPORT, ACProject.PATH_SEPARATOR);
    IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
    for (String filename : collection) {
      if (filename == null || filename.trim().length() == 0) {
        continue;
      }
      Path path = new Path(filename);
      if (root.exists(path)) {
        result.add("-I");
        String relativePath = filename;
        try {
          relativePath = getRelativeFSPathFromResource(acproject, filename, false, Messages.ValidationUtils_import_folder_not_found);
        } catch (Exception e) {
          // ignore, not a workspace path
        }
        result.add(relativePath);
      } else if (new File(filename).isAbsolute()) {
        IPath location = acproject.getProject().getLocation();
        String relativePath = filename;
        if (location != null) {
          File projectLocation = location.toFile().getAbsoluteFile();
          relativePath = createRelativePath(projectLocation, new File(filename));
        }
        result.add("-I");
        result.add(relativePath);
      } else {
        if (!ignoreErrors)
          throw new IOException(NLS.bind(Messages.ValidationUtils_import_folder_not_found, filename));
      }
    }
    // add other options
//    for (String key : keys) {
//
//      addOption(store, result, key);
//
//    }

    getParameters(store, result, validate);
    //Force in the -syntaxcheck flag if the user has asked us to just validate
    //the input files.
//    if ( validate ) result.add("-syntaxcheck");
    return result;
  }


  
  //For each value in values, add param to params (if not null), 
  //followed by value.
  private static void addEachStringParam(String param, 
          Collection<String> values, Collection<String> params ) {
      for (String value : values) {
          if ( param != null ) params.add(param);
          params.add(value);        
    }
  }
  
  //If value is non-empty, add it to params
  private static void addIfNotEmpty(String value, Collection<String> params) {
      if ( value.length() > 0 ) params.add(value);
  }
  
  //If value is non empty, add param, and then value, to params
  private static void addStringParamIfNotEmpty(String param, String value, Collection<String> params) 
  {
      if ( value.length() > 0 ) {
          params.add(param);
          params.add(value);
      }
  }
  
  /**
   * Add a parameter and its string value to params.  The logic is as follows:
   * If the store has a preference named key + "-check" which is set to true
   *    if a preference named key has a value V
   *        add param, then add the preference value V
   *    otherwise, if valueIsOptional
   *        add param
   *
   * @param param   The command line parameter
   * @param key     The key under which the associated string value is stored
   * @param valueIsOptional true if param should still be used even if the 
   *    associated string value is empty
   * @param store   The preference store to look in
   * @param params  The collection of parameters to add to
   */
  private static void addCheckedStringParam(String param, String key, 
          boolean valueIsOptional, 
          IPreferenceStore store, Collection<String> params )  
  {
      if ( store.getBoolean(key + "-check") ) 
      {
          String value = store.getString(key);
          if ( value.length() > 0 ) {
              params.add(param);
              params.add(value );
          }
          else if ( valueIsOptional ) {
              params.add(param);
          }
      }
  }
  
  /**
   * 
   * @param store
   * @param params
   * @param validate Ignore language preference and force "-syntaxcheck";
   */
  private static void getParameters(IPreferenceStore store, 
          Collection<String> params, boolean validate)
  {    
      //language options
      Language lang;
      if ( validate ) lang = Language.LANG_VALIDATE;
      else lang = Language.valueOfPref( 
              store.getString( ASN1CPreferenceConstants.LANGUAGE ) );
      
      if ( lang == Language.LANG_JAVA ) {
          //decide on -java, -java4, or -cldc
          if ( store.getBoolean( ASN1CPreferenceConstants.JAVA_SRC14 ) )
              params.add("-java4");
          else if ( store.getBoolean( ASN1CPreferenceConstants.JAVA_CLDC ) )
              params.add("-cldc");
          else 
              params.add("-java");
      }
      else if ( lang == Language.LANG_CPP ) {
         if ( store.getBoolean( ASN1CPreferenceConstants.C_CPP11))
            params.add("-cpp11");
         else
            params.add("-cpp");            
      }
      else params.add( lang.getParam() );

      //basic options
      addIfNotEmpty( store.getString(ASN1CPreferenceConstants.USER_OPTIONS), params);
      
      
      String asnstd = store.getString(ASN1CPreferenceConstants.ASNSTD);
      if ( asnstd.length() > 0 ) {
          params.add( "-asnstd" );
          params.add( asnstd );
      }
      
      addCheckedStringParam("-asn1", ASN1CPreferenceConstants.GEN_ASN1, 
              true, store, params);
      
      addCheckedStringParam("-html", ASN1CPreferenceConstants.GEN_HTML, 
              true, store, params);

      if ( store.getBoolean(ASN1CPreferenceConstants.XSD)) params.add("-xsd");
      
      if ( store.getBoolean(ASN1CPreferenceConstants.LAX_SYNTAX ) ) 
          params.add("-laxsyntax");

      
      addStringParamIfNotEmpty("-config", store.getString(ASN1CPreferenceConstants.CONFIG), params );
      
      if ( store.getBoolean(ASN1CPreferenceConstants.DEPENDS ) ) 
          params.add("-depends");
            
      if ( store.getBoolean(ASN1CPreferenceConstants.NO_AMBIG_TAG ) ) 
         params.add("-noAmbigTag");

      if ( store.getBoolean(ASN1CPreferenceConstants.STRICT_SIZE ) ) 
         params.add("-strictSize");

      if ( store.getBoolean(ASN1CPreferenceConstants.LIST ) ) 
          params.add("-list");      
      
      Collection<String> pdus = ACProject.toArray( 
              store.getString(ASN1CPreferenceConstants.PDU), File.pathSeparator);
      addEachStringParam("-pdu", pdus, params);
      
      if ( store.getBoolean(ASN1CPreferenceConstants.WARNINGS ) ) 
          params.add("-warnings");      
      

      if ( lang != Language.LANG_VALIDATE ) {
          //Options that are for any real target language
          
          //encoding options
          if ( store.getBoolean( ASN1CPreferenceConstants.BER )) params.add("-ber");
          if ( store.getBoolean( ASN1CPreferenceConstants.CER)) params.add("-cer");
          if ( store.getBoolean( ASN1CPreferenceConstants.DER)) params.add("-der");
          if ( store.getBoolean( ASN1CPreferenceConstants.APER)) params.add("-aper");
          if ( store.getBoolean( ASN1CPreferenceConstants.UPER)) params.add("-uper");
          if ( store.getBoolean( ASN1CPreferenceConstants.XER)) params.add("-xer");
          if ( store.getBoolean( ASN1CPreferenceConstants.XML)) params.add("-xml");
          if ( store.getBoolean( ASN1CPreferenceConstants.MDER)) params.add("-mder");
          if ( store.getBoolean( ASN1CPreferenceConstants.JSON)) params.add("-json");
          if ( lang == Language.LANG_C && store.getBoolean( ASN1CPreferenceConstants.OER))
              params.add("-oer");
          
          if ( store.getBoolean(ASN1CPreferenceConstants.COMPACT ) ) { 
              params.add("-compact");
          }

          if ( store.getBoolean(ASN1CPreferenceConstants.COMPAT_CHECKED ) ) { 
              params.add("-compat");  
              params.add( store.getString(ASN1CPreferenceConstants.COMPAT) );
          }   
          
          if ( store.getBoolean(ASN1CPreferenceConstants.LAX ) ) 
              params.add("-lax");      

          if ( !store.getBoolean(ASN1CPreferenceConstants.NODECODE ) ) 
              params.add("-nodecode");      
          
          if ( !store.getBoolean(ASN1CPreferenceConstants.NOENCODE ) ) 
              params.add("-noencode");      
          
          if ( store.getBoolean(ASN1CPreferenceConstants.NO_INDEF_LEN ) ) 
              params.add("-noIndefLen");      

          if ( store.getBoolean(ASN1CPreferenceConstants.NO_OBJ_TYPES ) ) 
              params.add("-noObjectTypes");      

          if ( store.getBoolean(ASN1CPreferenceConstants.NO_OPEN_EXT ) ) 
              params.add("-noOpenExt");      

          boolean print = false;
          if ( store.getBoolean( ASN1CPreferenceConstants.PRINT_CHECKED ) ) {
              params.add("-print");
              print = true;
          }
          else if ( lang == Language.LANG_C || lang == Language.LANG_CPP ) {
              if ( store.getBoolean( ASN1CPreferenceConstants.PRO_PRT2STR ) ) {
                  params.add("-prtToStr");
                  print = true;
              }
              else if ( store.getBoolean(ASN1CPreferenceConstants.PRO_PRT2STRM) ) {
                  params.add("-prtToStrm");
                  print = true;
              }
          }
          
          if ( print && (lang == Language.LANG_C || lang == Language.LANG_CPP))
          {
              //name of output file for print methods
              addIfNotEmpty( store.getString(ASN1CPreferenceConstants.PRINT), params);
          
              addStringParamIfNotEmpty("-prtfmt", store.getString(ASN1CPreferenceConstants.PRINT_FMT), params);
          }  
          
          if ( store.getBoolean(ASN1CPreferenceConstants.SHORT_NAMES ) ) 
              params.add("-shortnames");      

          if ( store.getBoolean(ASN1CPreferenceConstants.TRACE ) ) 
              params.add("-trace");      

          if ( store.getBoolean(ASN1CPreferenceConstants.NO_UNIQUE_NAMES ) ) 
              params.add("-noUniqueNames");      
          
          if ( store.getBoolean(ASN1CPreferenceConstants.NO_DATE_STAMP ) ) 
              params.add("-nodatestamp");      
          
          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_COMPARE_CHECKED ) ) {
              params.add("-compare"); 
              if ( lang == Language.LANG_C || lang == Language.LANG_CPP) {
                  //name of file to hold generated comparison code
                  params.add(store.getString(ASN1CPreferenceConstants.PRO_COMPARE) );
              }
          }

          // Pro options

          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_EVENTS ) ) 
              params.add("-events");       

          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_STREAM ) ) 
              params.add("-stream");  
          
          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_TABLES ) ) 
              params.add("-tables"); 
          
          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_STRICT ) ) 
              params.add("-strict");
          
          addStringParamIfNotEmpty("-param", store.getString(ASN1CPreferenceConstants.PRO_PARAM), params );      

          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_READER ) ) 
              params.add("-reader"); 

          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_WRITER ) ) 
              params.add("-writer");           
      }

      if ( lang == Language.LANG_C || lang == Language.LANG_CPP ) {
          if ( store.getBoolean(ASN1CPreferenceConstants.NO_TYPES ) ) 
              params.add("-notypes"); 
          
          //C/C++ options
          if ( store.getBoolean(ASN1CPreferenceConstants.C_FULLY_QUAL_ENUMS ) ) 
              params.add("-fqenum");         

          if ( store.getBoolean(ASN1CPreferenceConstants.C_MAXLINES_CHECKED)) {
             addStringParamIfNotEmpty("-maxlines",
                     store.getString(ASN1CPreferenceConstants.C_MAXLINES), 
                     params );
          }
          else if ( store.getBoolean(ASN1CPreferenceConstants.PRO_MAXCFILES ) ) 
              params.add("-maxcfiles");         
          else if ( store.getBoolean(ASN1CPreferenceConstants.C_CFILE_CHECKED ) ) {
              addStringParamIfNotEmpty("-hfile", store.getString(ASN1CPreferenceConstants.C_HFILE), params);
              addStringParamIfNotEmpty("-cfile", store.getString(ASN1CPreferenceConstants.C_CFILE), params);         
          }
          
          if ( store.getBoolean(ASN1CPreferenceConstants.C_GEN_BIT_MACROS ) ) 
              params.add("-genBitMacros");         

          if ( store.getBoolean(ASN1CPreferenceConstants.C_GENFREE ) ) 
              params.add("-genFree");
          
          addCheckedStringParam("-hdrGuardPfx", ASN1CPreferenceConstants.C_HDR_GUARD,
                  false, store, params);

          addCheckedStringParam("-cppns", ASN1CPreferenceConstants.C_NAMESPACE,
                  false, store, params);
          
          if ( store.getBoolean(ASN1CPreferenceConstants.C_NO_ENUM_CONVERSION ) ) 
              params.add("-noEnumConvert");
          
          if ( !store.getBoolean(ASN1CPreferenceConstants.C_NOINIT ) ) 
              params.add("-noInit");         


          addStringParamIfNotEmpty("-oh", store.getString(ASN1CPreferenceConstants.C_OH), params );      
          addStringParamIfNotEmpty("-bindir", store.getString(ASN1CPreferenceConstants.C_BINDIR), params );
          addStringParamIfNotEmpty("-libdir", store.getString(ASN1CPreferenceConstants.C_LIBDIR), params );
          addStringParamIfNotEmpty("-objdir", store.getString(ASN1CPreferenceConstants.C_OBJDIR), params );
          addStringParamIfNotEmpty("-prjdir", store.getString(ASN1CPreferenceConstants.C_PRJDIR), params );

          if ( store.getBoolean(ASN1CPreferenceConstants.C_STATIC ) ) 
              params.add("-static");   

          if ( store.getBoolean(ASN1CPreferenceConstants.C_TABLE_UNIONS ) ) 
              params.add("-table-unions");   

          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_3GPP ) ) 
              params.add("-3gpp");         

          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_GENTEST_CHECKED ) ) { 
              params.add("-genTest");  
              params.add( store.getString(ASN1CPreferenceConstants.PRO_GENTEST) );

              if ( store.getBoolean(ASN1CPreferenceConstants.RAND_TEST ) ) 
                 params.add("-randtest");
          }
          
          if ( store.getBoolean(ASN1CPreferenceConstants.PRO_COPY_CHECKED ) ) {
              params.add("-copy");
              params.add( store.getString(ASN1CPreferenceConstants.PRO_COPY) );
          }     
          
          if ( store.getBoolean(ASN1CPreferenceConstants.C_GEN_3G_L3) ) {
             params.add("-3GL3");
          }

          if ( store.getBoolean(ASN1CPreferenceConstants.C_NO_BITSTR32) ) {
             params.add("-nobitstr32");
          }

          if ( store.getBoolean(ASN1CPreferenceConstants.C_USE_ENUM_TYPES) ) {
             params.add("-use-enum-types");
          }
          
          if ( store.getBoolean(ASN1CPreferenceConstants.C_PER_PAD_BITSTR)) {
             params.add("-perPadBitStrings");
          }

          // C/C++ makefile/project options
          addStringParamIfNotEmpty("-genMake", store.getString(ASN1CPreferenceConstants.CM_GENMAKE), params );

          if ( store.getBoolean(ASN1CPreferenceConstants.CM_VCPROJ ) ) { 
              params.add("-vcproj");
              params.add( store.getString(ASN1CPreferenceConstants.CM_VCPROJ_VER) );
          }
          
          if ( store.getBoolean(ASN1CPreferenceConstants.CM_DLL ) ) 
              params.add("-usedll");
          
          if ( store.getBoolean(ASN1CPreferenceConstants.CM_BUILD_DLL ) ) 
              params.add("-builddll");
          
          if ( store.getBoolean(ASN1CPreferenceConstants.CM_MT ) ) 
              params.add("-mt"); 
          
          if ( store.getBoolean(ASN1CPreferenceConstants.CM_W32 ) ) 
              params.add("-w32"); 
                    
          //Symbian options
          if ( store.getBoolean(ASN1CPreferenceConstants.SYMBIAN ) ) { 
              params.add("-symbian");   

              if ( store.getBoolean(ASN1CPreferenceConstants.SYMBIAN_DLL ) ) 
                  params.add("dll");
          }          
      }
      else if ( lang == Language.LANG_JAVA) {
    
          //Java options
          
          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_DIRS ) ) 
              params.add("-dirs");      
          
          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_GENBUILD ) ) 
              params.add("-genbuild");      
    
          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_GENANT ) ) 
              params.add("-genant");      
    
          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_GENJSOURCES ) ) 
              params.add("-genjsources");      
    
          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_GENMETA ) ) 
             params.add("-genmetadata");      

          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_GETSET ) ) 
              params.add("-getset");      

          if ( store.getBoolean(ASN1CPreferenceConstants.JAVA_NOEVENTS)) {
             params.add("-noevents");
          }
    
    
          addCheckedStringParam("-pkgpfx", ASN1CPreferenceConstants.JAVA_PKGPFX,
                  false, store, params);
          
          addCheckedStringParam("-pkgname", ASN1CPreferenceConstants.JAVA_PKGNAME,
                  false, store, params);
          
          addIfNotEmpty( 
                store.getString(ASN1CPreferenceConstants.JAVA_XMLPARSER),
                params);
      }
      else if ( lang == Language.LANG_C_SHARP) {
          // C# options
          if ( store.getBoolean(ASN1CPreferenceConstants.CSHARP_DIRS ) ) 
              params.add("-dirs");
    
          
          addCheckedStringParam("-nspfx", ASN1CPreferenceConstants.CSHARP_NSPFX,
                  false, store, params);
    
          addCheckedStringParam("-namespace", ASN1CPreferenceConstants.CSHARP_NAMESPACE,
                  false, store, params);
          
          if ( store.getBoolean(ASN1CPreferenceConstants.CSHARP_GENCS_SOURCES ) ) 
              params.add("-gencssources");
          
          if ( store.getBoolean(ASN1CPreferenceConstants.CSHARP_GENMETA ) ) 
             params.add("-genmetadata");      

          if ( store.getBoolean(ASN1CPreferenceConstants.CSHARP_OUTPUT_BY_MODULE)) {
              params.add("-csfile");
              params.add("*.cs");
          }
          else if ( store.getBoolean(ASN1CPreferenceConstants.CSHARP_OUTPUT_SINGLE_FILE)) {
              params.add("-csfile");
              params.add(store.getString(ASN1CPreferenceConstants.CSHARP_OUTPUT_SINGLE_FILE_NAME));
          }
          
          addIfNotEmpty( 
                store.getString(ASN1CPreferenceConstants.CSHARP_XMLPARSER),
                params);
      }

      // XSD options
      if ( store.getBoolean(ASN1CPreferenceConstants.XSD_APPINFO ) ) { 
          params.add("-appinfo");
          
          if ( store.getBoolean(ASN1CPreferenceConstants.XSD_APPINFO_TAGS)) {
              params.add("tags");
          }
          if ( store.getBoolean(ASN1CPreferenceConstants.XSD_APPINFO_ENUMERATIONS)) {
              params.add("enums");
          }
          if ( store.getBoolean(ASN1CPreferenceConstants.XSD_APPINFO_EXTENSIONS)) {
              params.add("exts");
          }          
      }
      
      if ( store.getBoolean(ASN1CPreferenceConstants.XSD_ATTRS ) ) { 
          params.add("-appinfo");
          
          if ( store.getBoolean(ASN1CPreferenceConstants.XSD_ATTRS_TAGS)) {
              params.add("tags");
          }
          if ( store.getBoolean(ASN1CPreferenceConstants.XSD_ATTRS_ENUMERATIONS)) {
              params.add("enums");
          }
          if ( store.getBoolean(ASN1CPreferenceConstants.XSD_ATTRS_EXTENSIONS)) {
              params.add("exts");
          }          
      }

      addStringParamIfNotEmpty("-targetns", store.getString(ASN1CPreferenceConstants.XSD_TARGETNS), params );

      addStringParamIfNotEmpty("-useAsn1Xsd", store.getString(ASN1CPreferenceConstants.XSD_USE_ASN1XSD), params );      

  }

  private static String getRelativeFSPathFromResource(ACProject acproject,
      String filename, boolean isFile, String errorMessage) throws IOException {
    IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
    final IResource resource;
    if (isFile) {
      resource = root.getFile(new Path(filename));
    } else {
      resource = root.getFolder(new Path(filename));
    }
    if (resource == null || !resource.exists()) {
      throw new IOException(NLS.bind(errorMessage, filename));
    }
    File f = resource.getLocation().toFile();
    IPath location = acproject.getProject().getLocation();
    String relativePath;
    if (location != null) {
      File projectLocation = location.toFile().getAbsoluteFile();
      relativePath = createRelativePath(projectLocation, f);
      return relativePath;
    }
    return filename;
  }
  public static String getCommandLineParameters(ACProject acproject, IPreferenceStore store, int lineWidth) {
    try {
      Collection<String> parameters = getParameters(acproject, store, null, false, true);
      StringBuilder sb = new StringBuilder();
      int lineStart = 0;
      for (String string : parameters) {
        //if the line would be longer that given width and the string to add
        //isn't too long for a single line anyway, insert a break
        if (sb.length() - lineStart + string.length()> lineWidth 
                && string.length() <= lineWidth ) 
        {
          sb.append('\n');
          lineStart = sb.length();
        }
        sb.append(' ').append(string);
      }
      return sb.toString();
    } catch (Exception e) {
      return e.toString();
    }
  }
  /**
     * Convert an absolute path to a relative path if it is under a given base directory.
     * @param basedir the base directory for relative paths
     * @param path the directory to resolve
     * @return the relative path
     * @throws IOException if canonical path fails
     */
  public static String createRelativePath(File basedir, File path)
      throws IOException {
    String canonicalBasedir = basedir.getCanonicalPath();
    String canonicalPath = path.getCanonicalPath();
    // equal
    if (canonicalPath.equals(canonicalBasedir)) {
      return "."; //$NON-NLS-1$
    }
    if (canonicalPath.startsWith(canonicalBasedir)) {
      if (canonicalPath.charAt(canonicalBasedir.length()) == File.separatorChar) {
        canonicalPath = canonicalPath.substring(canonicalBasedir.length() + 1);
      } else {
        canonicalPath = canonicalPath.substring(canonicalBasedir.length());
      }
    } else {
      int sz = Math.min(canonicalBasedir.length(), canonicalBasedir.length());
      int i;
      for (i = 0; i < sz; i++) {
        if (canonicalBasedir.charAt(i) != canonicalPath.charAt(i)) {
          break;
        }
      }
      if (i > 0) {
        int lastIndexOfSepChar = canonicalBasedir.substring(0, i).lastIndexOf(File.separatorChar);
        String str = ""; //$NON-NLS-1$
        String commonStr = canonicalBasedir.substring(lastIndexOfSepChar);
        for (int j = 0; j < commonStr.length(); j++) {
          if (commonStr.charAt(j) == File.separatorChar) {
            str += ".." + File.separator; //$NON-NLS-1$
          }
        }
        canonicalPath = str + canonicalPath.substring(lastIndexOfSepChar+1);
      }
    }
    return canonicalPath;
  }
}
