package com.objsys.asn1c.validator.preferences;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;

import net.sourceforge.asneditor.ui.preferences.common.AbstractConfigurationBlock;
import net.sourceforge.asneditor.ui.preferences.common.ITextButtonListener;
import net.sourceforge.asneditor.ui.preferences.common.OverlayPreferenceStore;
import net.sourceforge.asneditor.ui.preferences.common.ScrolledPageContent;
import net.sourceforge.asneditor.ui.preferences.common.TextButtonListener;
import net.sourceforge.asneditor.ui.preferences.common.OverlayPreferenceStore.OverlayKey;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.dialogs.ContainerSelectionDialog;
import org.eclipse.ui.dialogs.ResourceSelectionDialog;
import org.eclipse.ui.forms.widgets.ExpandableComposite;

import com.objsys.asn1c.Activator;
import com.objsys.asn1c.validator.ValidationUtils;
import com.objsys.asn1c.validator.preferences.OptionPreference.ComboOptionPreference;

/**
 * Includes logic for building "preference pages" (generate functions, java,
 * etc.), either for wizards or for a general preference page.  Includes
 * createControl method for creating the general preference page.
 *
 * Via the base class, manages the relationship between controls and
 * preferences, keeping them synchronized.
 *
 * The addXxxSection methods add the appropriate controls for the given section
 * to a given Composite.
 *
 * This class prevents controls from being enabled that should not be enabled,
 * according to the chosen target languages and xsd option.
 *
 * About AbstractConfigurationBlock (the base class, not documented elsewhere):
 *    AbstractConfigurationBlock's responsibility is to keep the UI controls
 *    synchronized with the preference values in the preference store.
 *
 *    ACB maintains mappings from preference key to control, and from controls
 *    to preference keys.
 * 
 *    ACB maintains dependency (master/slave) relationships, which allow
 *    slaves to be enabled/disabled when the master is selected/unselected.
 *
 */
public class ASN1CConfigurationBlock extends AbstractConfigurationBlock {

  private static final String OSYSLIC_TXT = "osyslic.txt"; //$NON-NLS-1$
  private static final String OSLICDIR = "OSLICDIR"; //$NON-NLS-1$

  private Label expiresLabel;
  private Label productLabel;
  private Text cmdLineText;
  private IPropertyChangeListener propertyChangeListener;
  
  /* The following ExpandableComposite objects are set by createControl,
   * which is only invoked when using this class for the plug-in preferences.
   */
  private ExpandableComposite cExcomposite;
  private ExpandableComposite cmExcomposite;
  private ExpandableComposite csharpExcomposite;
  private ExpandableComposite javaExcomposite;
  private ExpandableComposite xsdExcomposite;
  private ExpandableComposite basicExcomposite;
  private ExpandableComposite genFunctionsExcomposite;
  private ExpandableComposite spaceOptimizationExcomposite;
  private ExpandableComposite encodingsExcomposite;

  /**
   * The set of keys whose controls may be enabled.  Some may be disabled due
   * to exclusive dependencies, for example.
   */
  private Set<String> enabledKeys;
  
  /**
   * Each array in fExclusiveDependencies is a sect of controls, of which
   * at most one should be selected.
   */
  private List<Button[]> fExclusiveDependencies = new LinkedList<Button[]>();
  
  /**
   * Create an ASN1CConfigurationBlock, with current preference settings
   * drawn from the given preference store.  This allows use
   * of the configuration block with either project configuration or
   * plug-in preferences.
   * @param store
   */
  public ASN1CConfigurationBlock(OverlayPreferenceStore store) {
    super(store);
    enabledKeys = null;
  }

  protected OverlayPreferenceStore.OverlayKey[] createOverlayStoreKeys() {
    OverlayKey[] preferenceKeys = ASN1CPreferenceConstants.createOverlayStoreKeys();
    return preferenceKeys;
  }
/*
  public Control createControl_tabs(Composite parent) {
    super.createControl(parent);
    Composite p = createComposite(parent, 1);

    TabFolder tabFolder = new TabFolder(p, SWT.TOP);
    applyGridLayout(1, tabFolder);

    Composite composite;

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_general);
    addGeneralSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_basic);
    addBasicSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_encodings);
    addEncodingSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_java);
    addJavaSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_c);
    addCAndCppSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_cm);
    addCMakefileSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_csharp);
    addCSharpSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_pro);
    addProSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_xsd);
    addXSDSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_symbian);
    addSymbianSection(composite);

    composite = createTabItem(tabFolder, Messages.ASN1CConfigurationBlock_additional);
    addAdditionalSection(composite);
    return parent;
  }
*/
  public void addGeneralSection(Composite composite, boolean showCompilerPath) {
    Composite root = createComposite(composite, 1);
    productLabel= new Label(root, SWT.NONE);
    GridData gd= new GridData(GridData.FILL, GridData.BEGINNING, true, false);
    productLabel.setLayoutData(gd);

    expiresLabel= new Label(root, SWT.NONE);
    gd= new GridData(GridData.HORIZONTAL_ALIGN_FILL);
    expiresLabel.setLayoutData(gd);

    // updateLicenseStatus();

    if (showCompilerPath) {
      Composite local = createComposite(root, 3);
      ITextButtonListener filelistener = new TextButtonListener(){
        public void processPush() {
          FileDialog fileDialog = new FileDialog(Display.getCurrent().getActiveShell());
          fileDialog.setFileName(text.getText());
          String selectedFile = fileDialog.open();
          if (selectedFile != null) {
            text.setText(selectedFile);
            // updateLicenseStatus();
          }
        }
      };
      addLabelledTextFieldWithButton(local, ASN1CPreferenceConstants.COMPILER_PATH, Messages.ASN1CConfigurationBlock_Browse, filelistener, 30, false, -1);
      getPreferenceStore().addPropertyChangeListener(new IPropertyChangeListener(){
        public void propertyChange(PropertyChangeEvent event) {
          if (ASN1CPreferenceConstants.COMPILER_PATH.equals(event.getProperty())) {
            // updateLicenseStatus();
          }
        }});

      Label previewLabel= new Label(root, SWT.NONE);
      previewLabel.setText(Messages.ASN1CConfigurationBlock_preview);
      gd= new GridData(GridData.HORIZONTAL_ALIGN_FILL);
      previewLabel.setLayoutData(gd);

      cmdLineText = new Text(root, SWT.MULTI | SWT.BORDER | SWT.V_SCROLL);
      cmdLineText.setEditable(false);
      gd= new GridData(GridData.HORIZONTAL_ALIGN_FILL | GridData.VERTICAL_ALIGN_FILL);
      gd.heightHint = 60;
      cmdLineText.setLayoutData(gd);
      updateCmdLineStatus();
    }
  }
  @Override
  public void initialize() {
    super.initialize();
    propertyChanged();
  }

  /* Invoked by initialize and whenever a preference is changed in the
   * preference store.   */
  private void propertyChanged() {
    updateCmdLineStatus();
    updateLanguageSections();
  }

  /**
   * Enable/disable pages and keys according to the target language and XSD
   * generation option in the current preference store.
   */
  private void updateLanguageSections() {
    boolean enableXSD = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.XSD);
    String language = getPreferenceStore().getString(ASN1CPreferenceConstants.LANGUAGE);
    if (ASN1CPreferenceConstants.LANG_C.equals(language)) {
      enablePagesAndKeys(true, false, false, enableXSD);
    } else if (ASN1CPreferenceConstants.LANG_CPP.equals(language)) {
      enablePagesAndKeys(true, false, false, enableXSD);
    } else if (ASN1CPreferenceConstants.LANG_C_SHARP.equals(language)) {
      enablePagesAndKeys(false, true, false, enableXSD);
    } else if (ASN1CPreferenceConstants.LANG_JAVA.equals(language)) {
      enablePagesAndKeys(false, false, true, enableXSD);
    } else {
      enablePagesAndKeys(false, false, false, false);
    }
  }

  /**
   * Enable expandable composites according to the chosen target language
   * and whether or not XSD generation has been chosen.
   *
   * Enable controls associated with keys that should be enabled under the
   * given language.
   * 
   * @param cpp
   * @param csharp
   * @param java
   * @param xsd
   */
  private void enablePagesAndKeys(boolean cpp, boolean csharp, boolean java, boolean xsd) {
    final boolean generateEnabled = cpp | csharp | java;
    //Note: if createControl has not been called, all of the enable
    //and collapseIfDisabled calls could be skipped.     
    enable(encodingsExcomposite, generateEnabled);
    enable(basicExcomposite, generateEnabled);
    enable(genFunctionsExcomposite, generateEnabled);
    enable(spaceOptimizationExcomposite, generateEnabled);
    enable(cExcomposite, cpp);
    enable(cmExcomposite, cpp);
    enable(csharpExcomposite, csharp);
    enable(javaExcomposite, java);
    enable(xsdExcomposite, generateEnabled && xsd);

    collapseIfDisabled(encodingsExcomposite);
    collapseIfDisabled(basicExcomposite);
    collapseIfDisabled(genFunctionsExcomposite);
    collapseIfDisabled(spaceOptimizationExcomposite);
    collapseIfDisabled(cExcomposite);
    collapseIfDisabled(cmExcomposite);
    collapseIfDisabled(csharpExcomposite);
    collapseIfDisabled(javaExcomposite);
    collapseIfDisabled(xsdExcomposite);    
    enableKeys(cpp, csharp, java, xsd);
  }

  /**
   * Enable keys according to chosen target language.
   * This also adjusts the label for the print option.
   * This also sets the -genmake option for C and C++
   *
   * @param cpp
   * @param csharp
   * @param java
   * @param xsd
   */
   public void enableKeys(boolean cpp, boolean csharp, boolean java, boolean xsd) {
    Collection<String> preferenceKeys = new HashSet<String>();
    preferenceKeys.addAll(ASN1CPreferenceConstants.getAlwaysEnabledKeys());
    if (cpp) {
      preferenceKeys.addAll(ASN1CPreferenceConstants.getCppPreferenceKeys(xsd, true));
    } else if (csharp) {
      preferenceKeys.addAll(ASN1CPreferenceConstants.getCSharpPreferenceKeys(xsd));
    } else if (java) {
      preferenceKeys.addAll(ASN1CPreferenceConstants.getJavaPreferenceKeys(xsd));
    } else {
      preferenceKeys.addAll(ASN1CPreferenceConstants.getValidationPreferenceKeys());
    }
    setEnabledKeys(preferenceKeys);

    //label the print option correctly
    //If the control is created already, change the label on it.
    //In case the control is destroyed and recreated (wizard), update the label
    //in the OptionPreference.
    Button print = (Button) getControlForPreferenceKey(
                              ASN1CPreferenceConstants.PRINT_CHECKED );
    OptionPreference printOpt = ASN1CPreferenceConstants.getCommandLineOption(
                                 ASN1CPreferenceConstants.PRINT_CHECKED);
    if ( java | csharp ) {
      //relabel the print option to "standard"
      if ( print != null ) {
         String oldText = print.getText();
         if ( !oldText.equals(Messages.ASN1CConfigurationBlock_printStd)) {
            print.setText(Messages.ASN1CConfigurationBlock_printStd);
            //make sure the new label is visible
            print.getParent().layout();
         }
      }
      printOpt.setLabel(Messages.ASN1CConfigurationBlock_printStd);
   }
   else {
      //relabel the print option to "stdout"
      if ( print != null )
         print.setText(Messages.ASN1CConfigurationBlock_print);
      printOpt.setLabel(Messages.ASN1CConfigurationBlock_print);
   }
   if ( cpp ) {
      String genMake = fStore.getString(ASN1CPreferenceConstants.CM_GENMAKE);
      if ( genMake.length() == 0 ) {
         fStore.setValue(ASN1CPreferenceConstants.CM_GENMAKE, "makefile");
         Text makefile = (Text) getControlForPreferenceKey(
                                 ASN1CPreferenceConstants.CM_GENMAKE );
         if ( makefile != null ) makefile.setText("makefile");
      }
   }
 }
 
  /**
   * Enable/disable the given control.  If the control is not allowed to
   * be enabled, it is forced to being disabled.
   */
  protected void enableControl(Control control, boolean enabled) {
    if (control == null) {
      return;
    }
    if (enabled && enabledKeys != null) {
      String preferenceKeyForControl = getPreferenceKeyForControl(control);
      if (!enabledKeys.contains(preferenceKeyForControl)) {
        enabled = false;
      }
    }
    enable(control, enabled);
  }

  /**
   * Enable/disable the given control.
   * @param control If null, do nothing.
   * @param enabled
   */
   private void enable(Control control, boolean enabled) {
    if (control == null)
      return;
    control.setEnabled(enabled);
//    control.setVisible(enabled);
  }

  private void collapseIfDisabled(ExpandableComposite excomposite) {
    if (excomposite == null) {
      return;
    }
    if (!excomposite.isEnabled()) {
      excomposite.setExpanded(false);
      expandedStateChanged(excomposite);
    }
  }

  public void addGeneratedFunctionsSection(Composite composite) {
    Composite root = createComposite(composite, 4, true, 1);

    addOption(root, ASN1CPreferenceConstants.NOENCODE); // all
    addOption(root, ASN1CPreferenceConstants.JAVA_GETSET);// get/set J
    addOption(root, ASN1CPreferenceConstants.C_NOINIT);// initialization c
    addOption(root, ASN1CPreferenceConstants.PRO_STREAM);// Stream J
    //    Button derButton= (Button) getControlForPreferenceKey(ASN1CPreferenceConstants.DER);
    //    if (derButton != null) {
    //    //I don't make this an exclusive dependency because if the PER
    //    //button is disabled by the stream button being checked, it may not be
    //    //obvious to the user how to enable it (uncheck stream).
    //    createDependency(perButton, streamButton);
    //    }

    addOption(root, ASN1CPreferenceConstants.NODECODE); // all
    addOption(root, ASN1CPreferenceConstants.C_GEN_BIT_MACROS);// named bit macros c
    addOption(root, ASN1CPreferenceConstants.C_GENFREE);// mem free c

    Composite compact = createComposite(composite, 2, true, 1);
    // compare J + C#, C with textfield
    Composite local = createLocalComposite(compact);
    Button compareButton = (Button) addOption(local, ASN1CPreferenceConstants.PRO_COMPARE_CHECKED);
    Control[] compareControls = (Control[]) addOption(local, ASN1CPreferenceConstants.PRO_COMPARE);
    createDependency(compareButton, compareControls);

    // copy C with textfield
    local = createLocalComposite(compact);
    Button copyButton = (Button) addOption(local, ASN1CPreferenceConstants.PRO_COPY_CHECKED);
    Control[] copyControls = (Control[]) addOption(local, ASN1CPreferenceConstants.PRO_COPY);
    createDependency(copyButton, copyControls);

    // test J
    local = createLocalComposite(compact);
    Button genTestButton = (Button) addOption(local, ASN1CPreferenceConstants.PRO_GENTEST_CHECKED);// gentest
    Control[] genTestControls = (Control[]) addOption(local, ASN1CPreferenceConstants.PRO_GENTEST);
    createDependency(genTestButton, genTestControls);
    Button randTest = (Button) addOption(local, ASN1CPreferenceConstants.RAND_TEST);
    createDependency(genTestButton, randTest);
    // print to string/stream/stdout
    local = createGroup(compact, 4, true, Messages.ASN1CConfigurationBlock_print_group, 2);
    Button printNoneButton = (Button) addOption(local, ASN1CPreferenceConstants.PRINT_NONE);// print
    addOption(local, ASN1CPreferenceConstants.PRINT_CHECKED);// print
    addOption(local, ASN1CPreferenceConstants.PRO_PRT2STR); // c
    addOption(local, ASN1CPreferenceConstants.PRO_PRT2STRM); // c
    Control[] printControls = (Control[]) addOption(local, ASN1CPreferenceConstants.PRINT, -1, 3);
    createInverseDependency(printNoneButton, printControls);
    Combo printFmtControls = (Combo) addOption(local, ASN1CPreferenceConstants.PRINT_FMT);
    createInverseDependency(printNoneButton, printFmtControls);
/*   KTB: I don't see the use of the following code.
     Commenting it seems to have no effect.  I expect that is because the
     controls are all afterward initialized from the preferences anyway.
     This code also seems fishy.
     For example, why isn't the printButton selected
     if the PRINT preference is chosen and the language is C++? 
     
    boolean p = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.PRINT)
      && ASN1CPreferenceConstants.isJavaOrCSharpLanguage(getPreferenceStore());
    boolean p2str = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.PRO_PRT2STR)
      && ASN1CPreferenceConstants.isCOrCppLanguage(getPreferenceStore());
    boolean p2strm = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.PRO_PRT2STRM)
      && ASN1CPreferenceConstants.isCOrCppLanguage(getPreferenceStore());
    if (p) {
      printButton.setSelection(true);
    } else if (p2str) {
      print2StrButton.setSelection(true);
    } else if (p2strm) {
      print2StrmButton.setSelection(true);
    } else {
      printNoneButton.setSelection(true);
      //try this...I think what happens is that above we check the
      //language and adjust the radio button accordingly, but then when
      //the control is initialized, each control's key's current value is
      //retrieved and the control is updated.  Thus, I can get the print to
      //stream selected even though the language is Java.  Probably the
      //gui should also set it to false when the language is changed...
      //it disables the option but lets it remain selected.
      //Eclipse lets you select multiple radio buttons.
      //getPreferenceStore().setValue(ASN1CPreferenceConstants.PRINT_NONE, true);
    }*/
  }

  private Composite createLocalComposite(Composite composite) {
    final int columns = 2;
    Composite local = new Composite(composite, SWT.NONE);
    GridLayout layout = new GridLayout(columns, false);
    layout.marginWidth = 0;
    layout.horizontalSpacing = 0;
    local.setLayout(layout);
    GridData data = new GridData(GridData.FILL, GridData.BEGINNING, true, false);
    local.setLayoutData(data);
    return local;
  }

  /**
   * @param controls
   * @return return the button from the Checkbox + String Option
   */
  private Button getButton(Control[] controls) {
    return (Button) controls[2];
  }

  public void addSpaceOptimizationSection(Composite composite) {
    addOption(composite, ASN1CPreferenceConstants.NO_INDEF_LEN);
    addOption(composite, ASN1CPreferenceConstants.NO_OPEN_EXT);
    addOption(composite, ASN1CPreferenceConstants.LAX);
    addOption(composite, ASN1CPreferenceConstants.COMPACT);
    addOption(composite, ASN1CPreferenceConstants.C_NO_ENUM_CONVERSION);

    Composite local = createLocalComposite(composite);
    Button compatButton = (Button) addOption(local, ASN1CPreferenceConstants.COMPAT_CHECKED);
    Combo compatCombo = (Combo) addOption(local, ASN1CPreferenceConstants.COMPAT, true);
    createDependency(compatButton, compatCombo);
  }

  /**
   * @param composite the gui parent
   * @param fromWizard <b>true</b> means this is called from the project configuration Wizard
   */
  public void addBasicSection(Composite composite, boolean fromWizard) {
    if (fromWizard)
      addButtonOption(composite, ASN1CPreferenceConstants.CONFIG,
      Messages.ASN1CConfigurationBlock_Browse, createFileListener());
    if (fromWizard)
      addOption(composite, ASN1CPreferenceConstants.IMPORT);
    addOption(composite, ASN1CPreferenceConstants.NO_OBJ_TYPES);
    addOption(composite, ASN1CPreferenceConstants.NO_TYPES);
    if (fromWizard)
      addButtonOption(composite, ASN1CPreferenceConstants.OUTPUT_DIR,
      Messages.ASN1CConfigurationBlock_Browse, createFolderListener());
    addOption(composite, ASN1CPreferenceConstants.PDU);
    addOption(composite, ASN1CPreferenceConstants.SHORT_NAMES);
    addOption(composite, ASN1CPreferenceConstants.TRACE);
    addOption(composite, ASN1CPreferenceConstants.NO_UNIQUE_NAMES);
    if (fromWizard) {
      addOption(composite, ASN1CPreferenceConstants.LIST);
      addOption(composite, ASN1CPreferenceConstants.WARNINGS);
    }
    addOption(composite, ASN1CPreferenceConstants.NO_DATE_STAMP);

    addOption(composite, ASN1CPreferenceConstants.PRO_EVENTS);
    addOption(composite, ASN1CPreferenceConstants.PRO_TABLES);
    addOption(composite, ASN1CPreferenceConstants.PRO_STRICT);
    addOption(composite, ASN1CPreferenceConstants.STRICT_SIZE);
    addOption(composite, ASN1CPreferenceConstants.PRO_3GPP);
    addOption(composite, ASN1CPreferenceConstants.C_TABLE_UNIONS);
    //    Button per = (Button) getControlForPreferenceKey(
    //    		ASN1CPreferenceConstants.PER );
    //Defining the dependency causes the 3GPP checkbox to be indented.
    //I grab the GridData and fix this.
    //    createDependency(per, _3gpp);
    //    ((GridData)(_3gpp.getLayoutData())).horizontalIndent = 0;
    addOption(composite, ASN1CPreferenceConstants.PRO_PARAM);
    addOption(composite, ASN1CPreferenceConstants.PRO_READER);
    addOption(composite, ASN1CPreferenceConstants.PRO_WRITER);

    addOption(composite, ASN1CPreferenceConstants.USER_OPTIONS);
  }

  public void addLanguageSection(Composite composite) {
    addOption(composite, ASN1CPreferenceConstants.LANGUAGE);
    addOption(composite, ASN1CPreferenceConstants.ASNSTD);
    addOption(composite, ASN1CPreferenceConstants.DEPENDS);
    addOption(composite, ASN1CPreferenceConstants.XSD);
    addOption(composite, ASN1CPreferenceConstants.GEN_ASN1);
    addOption(composite, ASN1CPreferenceConstants.GEN_HTML);
    addOption(composite, ASN1CPreferenceConstants.LAX_SYNTAX);
    addOption(composite, ASN1CPreferenceConstants.NO_AMBIG_TAG);
  }

  public Control[] addButtonOption(Composite composite, String key, String buttonLabel, ITextButtonListener listener) {
    Composite local = createComposite(composite, 3);
    return addLabelledTextFieldWithButton(local, key, buttonLabel, listener, 30, false, -1);
  }

  public void addEncodingSection(Composite composite) {
    Composite local = createComposite(composite, 6);
    Button ber = (Button) addOption(local, ASN1CPreferenceConstants.BER);
    Button cer = (Button) addOption(local, ASN1CPreferenceConstants.CER);
    Button der = (Button) addOption(local, ASN1CPreferenceConstants.DER);
    Button aper = (Button) addOption(local, ASN1CPreferenceConstants.APER);
    Button uper = (Button) addOption(local, ASN1CPreferenceConstants.UPER);
    Button xer = (Button) addOption(local, ASN1CPreferenceConstants.XER);
    Button xml = (Button) addOption(local, ASN1CPreferenceConstants.XML);
    addOption(local, ASN1CPreferenceConstants.OER);
    addOption(local, ASN1CPreferenceConstants.MDER);
    addOption(local, ASN1CPreferenceConstants.JSON);    

    Button streamButton = (Button) getControlForPreferenceKey(
                                       ASN1CPreferenceConstants.PRO_STREAM);
    if (streamButton != null)
      //I don't make this an exclusive dependency because if the PER
      //button is disabled by the stream button being checked, it may not be
    	//obvious to the user how to enable it (uncheck stream).
      createDependency(aper, streamButton); 
      
    Button[] buttons = new Button[]{ber, cer, der};
    /*
     * KTB: The following code seems to serve no purpose.
     * The values are all initialized elsewhere.
    boolean b = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.BER);
    boolean c = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.CER);
    boolean d = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.DER);
    if (b) {
      ber.setSelection(true);
      cer.setSelection(false);
      der.setSelection(false);
    } else if (c) {
      ber.setSelection(false);
      cer.setSelection(true);
      der.setSelection(false);
    } else if (d) {
      ber.setSelection(false);
      cer.setSelection(false);
      der.setSelection(true);
    } else {
      ber.setSelection(false);
      cer.setSelection(false);
      der.setSelection(false);
    }
    exclusiveSelectedEvent(buttons, b ? ber : c ? cer : der);*/
    createExclusiveDependency(buttons);

    buttons = new Button[]{xml, xer};
    /*
     * KTB: The following code seems to serve no purpose.
     * The values are all initialized elsewhere.
    boolean xm = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.XML);
    boolean xr = getPreferenceStore().getBoolean(ASN1CPreferenceConstants.XER);
    if (xm) {
      xml.setSelection(true);
      xer.setSelection(false);
    } else if (xr) {
      xml.setSelection(false);
      xer.setSelection(true);
    } else {
      xml.setSelection(false);
      xer.setSelection(false);
    }
    exclusiveSelectedEvent(buttons, xm ? xml : xer);*/
    createExclusiveDependency(buttons);
  }
  
  /**   
   * Define a set of (check box) controls as being exclusive: at most one of
   * the controls should ever be selected and enabled at the same time.
   * If none are selected, all may be enabled.  If one is selected, all others
   * are disabled.  This is used for BER/CER/DER selection, and XML/XER
   * selection.
   * This is similar to radio buttons, except if radio buttons were used we
   * would need a "none" option.
   *
   * Creates a selection listener to enable/disable related controls.
   * Adds array to fExclusiveDependencies so that updateSlaves can include
   * updating exclusive dependencies.
   *
   * @param controls
   */
  private void createExclusiveDependency(final Button[] controls) {
    SelectionListener listener= new SelectionListener() {
      boolean running = false;
      public void widgetSelected(SelectionEvent e) {
        if (running) {
          return;
        }
        running = true;
        Button source = (Button) e.getSource();
        exclusiveSelectedEvent(controls, source);
        running = false;
      }
      public void widgetDefaultSelected(SelectionEvent e) {}
    };
    for (Button button : controls) {
      button.addSelectionListener(listener);
    }
    fExclusiveDependencies.add(controls);
  }

  public void addJavaSection(Composite composite) {
    addOption(composite, ASN1CPreferenceConstants.JAVA_GENBUILD);
    addOption(composite, ASN1CPreferenceConstants.JAVA_GENANT);
    addOption(composite, ASN1CPreferenceConstants.JAVA_GENJSOURCES);
    addOption(composite, ASN1CPreferenceConstants.JAVA_GENMETA);
    addOption(composite, ASN1CPreferenceConstants.JAVA_SRC14);
    addOption(composite, ASN1CPreferenceConstants.JAVA_CLDC);

    addOption(composite, ASN1CPreferenceConstants.JAVA_DIRS);
    addExclusiveOptions(composite, ASN1CPreferenceConstants.JAVA_PKGNAME,
                         ASN1CPreferenceConstants.JAVA_PKGPFX);
    addOption(composite, ASN1CPreferenceConstants.JAVA_NOEVENTS);
    addOption(composite, ASN1CPreferenceConstants.JAVA_XMLPARSER);
  }

  private void addExclusiveOptions(Composite composite, String... keys) {
    if (keys == null || keys.length == 0) {
      return;
    }
    Control[][] controls = new Control[keys.length][];
    Button[] buttons = new Button[keys.length];
    boolean[] pref = new boolean[keys.length];

    for (int i = 0; i < keys.length; i++) {
      controls[i] = (Control[]) addOption(composite, keys[i], true);
      buttons[i] = getButton(controls[i]);
      pref[i] = getPreferenceStore().getBoolean(getEnableCheckboxKeyFor(keys[i]));
    }

    for (int i = 0; i < pref.length; i++) {
      if (!pref[i]) {
        continue;
      }
      for (int j = 0; j < buttons.length; j++) {
        buttons[j].setSelection(i==j);
      }
      exclusiveSelectedEvent(buttons, buttons[i]);
      break;
    }
    createExclusiveDependency(buttons);
  }

  public void addCAndCppSection(Composite composite) {
    addOption(composite, ASN1CPreferenceConstants.C_CPP11);     
    
    Composite local = createGroup(composite, 3, Messages.ASN1CConfigurationBlock_outputFiles);
    Button maxLinesButton = (Button) addOption(local, ASN1CPreferenceConstants.C_MAXLINES_CHECKED);
    Control[] maxLinesControls = (Control[]) addOption(local, ASN1CPreferenceConstants.C_MAXLINES, -1, 2);
    createDependency(maxLinesButton, maxLinesControls);

    addOption(local, ASN1CPreferenceConstants.PRO_MAXCFILES, -1, 3);

    Button cfileButton = (Button) addOption(local, ASN1CPreferenceConstants.C_CFILE_CHECKED);
    Control[] chfileCtrl = (Control[]) addOption(local, ASN1CPreferenceConstants.C_HFILE);
    Control[] ccfileCtrl = (Control[]) addOption(local, ASN1CPreferenceConstants.C_CFILE);
    List<Control> cfileCtrl = new ArrayList<Control>(chfileCtrl.length + ccfileCtrl.length);
    cfileCtrl.addAll(Arrays.asList(chfileCtrl));
    cfileCtrl.addAll(Arrays.asList(ccfileCtrl));
    createDependency(cfileButton, cfileCtrl.toArray(new Control[cfileCtrl.size()]));
    
    addOption(composite, ASN1CPreferenceConstants.C_OH);
    addOption(composite, ASN1CPreferenceConstants.C_BINDIR);
    addOption(composite, ASN1CPreferenceConstants.C_LIBDIR);
    addOption(composite, ASN1CPreferenceConstants.C_OBJDIR);
    addOption(composite, ASN1CPreferenceConstants.C_PRJDIR);
    addOption(composite, ASN1CPreferenceConstants.C_STATIC);

    local = createLocalComposite(composite);
    Button symbian = (Button) addOption(local, ASN1CPreferenceConstants.SYMBIAN);
    Button dll = (Button) addOption(local, ASN1CPreferenceConstants.SYMBIAN_DLL);
    createDependency(symbian, dll);
    addOption(composite, ASN1CPreferenceConstants.C_FULLY_QUAL_ENUMS);
    
    //Since C_HDR_GUARD is defined as an optional string, it will automatically
    //have an enable check box created for it.  Use noMarginWidth = true to
    //avoid having the check box indented.  Likewise for C_NAMESPACE.
    addOption(composite, ASN1CPreferenceConstants.C_HDR_GUARD, true);
    addOption(composite, ASN1CPreferenceConstants.C_NAMESPACE, true);
    addOption(composite, ASN1CPreferenceConstants.C_GEN_3G_L3);
    addOption(composite, ASN1CPreferenceConstants.C_NO_BITSTR32);
    addOption(composite, ASN1CPreferenceConstants.C_USE_ENUM_TYPES);
    addOption(composite, ASN1CPreferenceConstants.C_PER_PAD_BITSTR);
  }

  public void addCMakefileSection(Composite composite) {
    addOption(composite, ASN1CPreferenceConstants.CM_GENMAKE);
    Button vcProj = (Button) addOption(composite,
                                       ASN1CPreferenceConstants.CM_VCPROJ);
    Control vcProjVer = (Control) addOption(composite,
                                       ASN1CPreferenceConstants.CM_VCPROJ_VER);
    createDependency(vcProj, vcProjVer);
    addOption(composite, ASN1CPreferenceConstants.CM_BUILD_DLL);
    addOption(composite, ASN1CPreferenceConstants.CM_DLL);
    addOption(composite, ASN1CPreferenceConstants.CM_MT);
    addOption(composite, ASN1CPreferenceConstants.CM_W32);
  }

  public void addCSharpSection(Composite composite) {
    //Use 3 columns, as the single file option has the radio button, the field,
    //and the button for editing the field.
    //For the two buttons that standalone, set the column span to 3.  This
    //makes it so each radio button is on its own row.
    Composite outputGroup = createGroup(composite, /*columns=*/3, Messages.ASN1CConfigurationBlock_outputFiles);
    Button outputByClass = (Button) addOption(outputGroup, ASN1CPreferenceConstants.CSHARP_OUTPUT_BY_CLASS, -1, 3);
    addOption(outputGroup, ASN1CPreferenceConstants.CSHARP_OUTPUT_BY_MODULE, -1, 3);
    Button outputSingleFile = (Button) addOption(outputGroup, ASN1CPreferenceConstants.CSHARP_OUTPUT_SINGLE_FILE);
    Control[] singleFileCtrls = (Control[]) addOption(outputGroup, ASN1CPreferenceConstants.CSHARP_OUTPUT_SINGLE_FILE_NAME);

    //Don't include the label in the dependency?  Once disabled, it remains
    //grayed out forever.
    singleFileCtrls = new Control[] { singleFileCtrls[1], singleFileCtrls[2] };
    createDependency(outputSingleFile, singleFileCtrls );
    addOption(composite, ASN1CPreferenceConstants.CSHARP_GENCS_SOURCES);
    addOption(composite, ASN1CPreferenceConstants.CSHARP_GENMETA);
    addOption(composite, ASN1CPreferenceConstants.CSHARP_DIRS);
    addExclusiveOptions(composite, ASN1CPreferenceConstants.CSHARP_NAMESPACE, ASN1CPreferenceConstants.CSHARP_NSPFX);
    addOption(composite, ASN1CPreferenceConstants.CSHARP_XMLPARSER);
  }

  public void addXSDSection(Composite composite) {
    Button app = (Button)addOption(composite, ASN1CPreferenceConstants.XSD_APPINFO);
    Composite local = createComposite(composite, 3);
    Button app_tags = (Button)addOption(local, ASN1CPreferenceConstants.XSD_APPINFO_TAGS);
    Button app_enums = (Button)addOption(local, ASN1CPreferenceConstants.XSD_APPINFO_ENUMERATIONS);
    Button app_exts = (Button)addOption(local, ASN1CPreferenceConstants.XSD_APPINFO_EXTENSIONS);
    createDependency(app, new Control[]{app_tags, app_enums, app_exts});

    Button attr = (Button)addOption(composite, ASN1CPreferenceConstants.XSD_ATTRS);
    local = createComposite(composite, 3);
    Button attr_tags = (Button)addOption(local, ASN1CPreferenceConstants.XSD_ATTRS_TAGS);
    Button attr_enums = (Button)addOption(local, ASN1CPreferenceConstants.XSD_ATTRS_ENUMERATIONS);
    Button attr_exts = (Button)addOption(local, ASN1CPreferenceConstants.XSD_ATTRS_EXTENSIONS);
    createDependency(attr, new Control[]{attr_tags, attr_enums, attr_exts});

    addOption(composite, ASN1CPreferenceConstants.XSD_TARGETNS);
    addOption(composite, ASN1CPreferenceConstants.XSD_USE_ASN1XSD);
  }

  /**
   * Scan for the license file and set the values to the labels
   * @param productLabel
   * @param expiresLabel
   */
  /*
   * 13-01-08 MEM: the updateLicenseStatus function only works with osyslic
   * style licenses.  We don't use these anymore, and in any case the plugin
   * does not itself do any real license checking.  I've disabled this in 
   * response to ASN-5157.
   *
   * There are currently five different license templates available to us in 
   * win64:/cygdrive/d/acshared/RLM/license_templates/ASN1C that would need
   * to be parsed properly.  One of these is a client, which doesn't really 
   * give any product information at all.
   *
   * With these developments, I question the relevance of this method.
   */
  private void updateLicenseStatus() {
    if (productLabel == null || productLabel.isDisposed() ||
        expiresLabel == null || expiresLabel.isDisposed()) {
      return;
    }

    String product = null, version = null, expire = null;

    String oslicdir = System.getenv(OSLICDIR);
    if (oslicdir == null) {
      oslicdir = getPreferenceStore().getString(ASN1CPreferenceConstants.COMPILER_PATH);
      if (oslicdir != null) {
        File parentFile = new File(oslicdir).getParentFile();
        if (parentFile != null) {
          oslicdir = parentFile.getAbsolutePath();
        }
      }
    }
    if (oslicdir != null) {
      File licFile = new File(oslicdir, OSYSLIC_TXT);
      BufferedReader fileReader = null;
      try {
        fileReader = new BufferedReader(new FileReader(licFile));
        String line = null;
        while ((product == null || version == null || expire == null)
            && (line = fileReader.readLine()) != null) {
          if (line.startsWith("#")) { //$NON-NLS-1$
            continue;
          }
          if (line.startsWith("Product: ")) { //$NON-NLS-1$
            product = line.substring(line.indexOf(':')+1).trim();
            continue;
          }
          if (line.startsWith("Version: ")) { //$NON-NLS-1$
            version = line.substring(line.indexOf(':')+1).trim();
            continue;
          }
          if (line.startsWith("Expiration: ")) { //$NON-NLS-1$
            expire = line.substring(line.indexOf(':')+1).trim();
            continue;
          }
        }
      } catch (Exception e) {
        product = null;
        version = null;
        expire = null;
      } finally {
        try {
          if (fileReader != null)
            fileReader.close();
        } catch (Exception e) {
          /* ignore */
        }
      }
    }
    String licenseString = (product==null) ? Messages.ASN1CConfigurationBlock_no_license : product+" "+version; //$NON-NLS-1$
    productLabel.setText(licenseString);
    expiresLabel.setText(NLS.bind(Messages.ASN1CConfigurationBlock_Expire_date, expire == null ? Messages.ASN1CConfigurationBlock_expire_invalid : expire));
  }

  public ITextButtonListener createFileListener() {
    ITextButtonListener filelistener = new TextButtonListener(){
      public void processPush() {
        FileDialog fileDialog = new FileDialog(Display.getCurrent().getActiveShell());
        fileDialog.setFileName(text.getText());
        String selectedFile = fileDialog.open();
        if (selectedFile != null) {
          text.setText(selectedFile);
        }
      }
    };
    return filelistener;
  }

  public ITextButtonListener createFolderListener() {
    ITextButtonListener filelistener = new TextButtonListener(){
      public void processPush() {
        DirectoryDialog fileDialog = new DirectoryDialog(Display.getCurrent().getActiveShell());
        fileDialog.setFilterPath(text.getText());
        String selectedFile = fileDialog.open();
        if (selectedFile != null) {
          text.setText(selectedFile);
        }
      }
    };
    return filelistener;
  }

  /**
   * Uses the standard container selection dialog to choose the new value for
   * the container field.
   */
  public ITextButtonListener createWorkspaceFolderListener(IContainer root, final String message) {
    if (root == null) {
      root = ResourcesPlugin.getWorkspace().getRoot();
    }
    final IContainer internalRoot = root;

    ITextButtonListener filelistener = new TextButtonListener(){
      public void processPush() {
        IContainer initContainer = internalRoot;
        if (text.getText().trim().length() > 0) {
          try {
            initContainer = ResourcesPlugin.getWorkspace().getRoot().getFolder(new Path(text.getText()));
          } catch (Exception e) {
            initContainer = internalRoot;
          }
        }
        ContainerSelectionDialog dialog = new ContainerSelectionDialog(
            Display.getCurrent().getActiveShell(), initContainer, false, message);

        if (dialog.open() == ContainerSelectionDialog.OK) {
          Object[] result = dialog.getResult();
          if (result.length == 1) {
            Path path = ((Path) result[0]);
            String string = path.toPortableString();
            text.setText(string);
          }
        }
      }
    };
    return filelistener;
  }

  /**
   * Uses the standard container selection dialog to choose the new value for
   * the resource field.
   */
  public ITextButtonListener createWorkspaceResourceListener(IContainer root, final String message) {
    if (root == null) {
      root = ResourcesPlugin.getWorkspace().getRoot();
    }
    final IContainer internalRoot = root;

    ITextButtonListener filelistener = new TextButtonListener(){
      public void processPush() {
        IContainer initContainer = internalRoot;
        if (text.getText().trim().length() > 0) {
          try {
            String absolutePath = new File(text.getText()).getAbsolutePath();
            initContainer = ResourcesPlugin.getWorkspace().getRoot().getContainerForLocation(new Path(absolutePath));
          } catch (Exception e) {
            initContainer = internalRoot;
          }
        }
        ResourceSelectionDialog dialog = new ResourceSelectionDialog(
            Display.getCurrent().getActiveShell(), initContainer, message);

        if (dialog.open() == ContainerSelectionDialog.OK) {
          Object[] result = dialog.getResult();
          if (result.length >= 1) {
            Path path = ((Path) result[0]);
            IFolder folder = ResourcesPlugin.getWorkspace().getRoot().getFolder(path);
            String string = folder.getLocation().toOSString();
            text.setText(string);
          }
        }
      }
    };
    return filelistener;
  }

  private Object addOption(Composite composite, String key, boolean noMarginWidth) {
    return addOption(composite, key, -1, 1, noMarginWidth);
  }
  
  /**   
   * Add a control to composite for the given key.
   * The key should correspond to an OptionPreference in
   * ASN1CPreferenceConstants, and this OptionPreference will be used to decide
   * what kind of control should be added to the composite.
   *
   * @param composite
   * @param key
   * @return
   */
  public Object addOption(Composite composite, String key) {
    return addOption(composite, key, -1, 1);
  }
  protected Object addOption(Composite composite, String key, int indent, int colspan) {
    return addOption(composite, key, indent, colspan, false);
  }
  
  /** 
   * Add control to composite for the given preference key.
   *
   * This will lookup the OptionPreference for the given key (call it op).
   * If {@code op.isBooleanOption()}, create a radio button or a check box,
   * depending on {@code op.isRadioButton()}.
   * If {@code op.isComboOption()), create a combo box for it.
   * Otherwise, create a labeled text field, possibly with an "enable check
   * box".
   *
   * @param composite
   * @param key
   * @param indent
   * @param colspan
   * @param noMarginWidth
   * @return The control that was created.  In the case of a labeled text field,
   *     the controls that were created.
   */
  protected Object addOption(Composite composite, String key, int indent, int colspan, boolean noMarginWidth) {
    final OptionPreference option = ASN1CPreferenceConstants.getCommandLineOption(key);
    if (option.isBooleanOption()) {
      if (option.isRadioButton())
        return addRadioButton(composite, option.getLabel(), key, indent, colspan);
      else
        return addCheckBox(composite, option.getLabel(), key, indent, colspan);
    }
    if (option.isComboOption()) {
      Composite local = createComposite(composite, 2, colspan);
      Combo addComboBox = addComboBox(local, option.getLabel(), key,
          ((ComboOptionPreference)option).getValues(), ((ComboOptionPreference)option).getValueLabels(), indent);
      return addComboBox;
    }
    Composite local = createComposite(composite, 3, colspan);
    if (noMarginWidth) {
      ((GridLayout)local.getLayout()).marginWidth = 0;
    }
    return addLabelledTextFieldWithButton(local, option.getLabel(),
        Messages.ASN1CConfigurationBlock_Edit, key,
        new TextButtonListener() {
          public void processPush() {
            InputDialog in = new InputDialog(Display.getCurrent().getActiveShell(),
                option.getLabel(), option.getLabel(), text.getText(), null, true);
            in.setInitialSize(new Point(400, 300));
            if (InputDialog.OK == in.open()) {
              text.setText(in.getValue());
            }
          }
        }, 10, false, indent, option.isAddEnableCheckBox());
  }

  private Control[] addLabelledTextFieldWithButton(Composite composite, String key, String buttonLabel,
      ITextButtonListener listener, int textLimit, boolean hardLimit, int indentation) {
    return addLabelledTextFieldWithButton(composite, getLabel(key), buttonLabel, key, listener, textLimit, hardLimit, indentation);
  }

  private String getLabel(String key) {
    OptionPreference option = ASN1CPreferenceConstants.getCommandLineOption(key);
    return option.getLabel();
  }

  private void updateCmdLineStatus() {
    if (cmdLineText== null || cmdLineText.isDisposed()) {
      return;
    }
    try {
      String string = ValidationUtils.getCommandLineParameters(null, getPreferenceStore(), 70);
      cmdLineText.setText(string);
    } catch (Exception e) {
      e.printStackTrace();
      // ignore
    }
  }

  /**
   * Create a control with expandable composites for each
   * section of preferences.  This is for use with the ASN1C Plugin
   * preferences dialog (as opposed to the ASN1C Plugin Project configuration
   * dialog).
   *
   * This also sets up a change listener so that when a change is made,
   * we can invoke the propertyChanged method.
   *
   */
   public Control createControl(Composite parent) {
    ScrolledPageContent scrolled= new ScrolledPageContent(parent);
    scrolled.setExpandHorizontal(true);
    scrolled.setExpandVertical(true);

    Composite control= scrolled.getBody();
    GridLayout layout= new GridLayout();
    layout.marginWidth= 0;
    layout.marginHeight= 0;
    control.setLayout(layout);

    ExpandableComposite excomposite;
    Composite composite;

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_general, 1);
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    excomposite.setExpanded(true);
    addGeneralSection(composite, true);

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_language, 1);
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addLanguageSection(composite);

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_encodings, 1);
    encodingsExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addEncodingSection(composite);

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_genFunctions, 1);
    genFunctionsExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addGeneratedFunctionsSection(composite);

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_spaceOptimization, 1);
    spaceOptimizationExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addSpaceOptimizationSection(composite);

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_basic, 1);
    basicExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addBasicSection(composite, true);

    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_java, 1);
    javaExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addJavaSection(composite);


    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_c, 1);
    cExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addCAndCppSection(composite);


    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_cm, 1);
    cmExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addCMakefileSection(composite);


    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_csharp, 1);
    csharpExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addCSharpSection(composite);


    excomposite = createStyleSection(control, Messages.ASN1CConfigurationBlock_xsd, 1);
    xsdExcomposite = excomposite;
    composite = createComposite(excomposite, 1);
    excomposite.setClient(composite);
    addXSDSection(composite);

    scrolled.setContent(control);
//    final Point size= control.computeSize(SWT.DEFAULT, SWT.DEFAULT);
//    scrolled.setMinSize(size.x, size.y);
    //Install a property change listener on the preference store.  Invoke
    //propertyChanged whenever a property is changed.
    propertyChangeListener = new IPropertyChangeListener(){
      public void propertyChange(PropertyChangeEvent event) {
        propertyChanged();
      }
    };
    getPreferenceStore().addPropertyChangeListener(propertyChangeListener);
    updateLanguageSections();

    return scrolled;
  }

  public void dispose() {
    if (propertyChangeListener != null)
      getPreferenceStore().removePropertyChangeListener(propertyChangeListener);
    super.dispose();
  }
  
  /**
   * Handle selection of an exclusively-selectable set of controls.
   * If source is selected, unselect and disable controls (except source).
   * If source is not selected, and none of controls is selected, enable
   * all of controls.
   */
  private void exclusiveSelectedEvent(final Button[] controls, Button source) {
    boolean state= source.getSelection();
    if (state) {
      for (Button button : controls) {
        if (source != button) {
          button.setSelection(false);
          setControlEnable(button, false);
        }
      }
    } else {
      boolean active = isOneSelected(controls);
      if(!active) {
        for (Button button : controls) {
          setControlEnable(button, true);
        }
      }
    }
  }

  private static boolean isOneSelected(final Button[] controls) {
    boolean active = false;
    for (Button button : controls) {
      active |= button.getSelection();
    }
    return active;
  }

  /**
   * Return the first control that is selected from
   * amongst the given array of controls.  If no
   * control is selected, return null.
   *
   * @param controls
   * @return
   */
   private Button getSelected(final Button[] controls) {
    for (Button button : controls) {
      if (button.getSelection()) {
        return button;
      }
    }
    return null;
  }

  public void setControlEnable(Control control, boolean value) {
    if (value && enabledKeys != null) {
      String prefName = getPreferenceKeyForControl(control);
      if (prefName == null) {
        Activator.log(new Status(IStatus.WARNING, Activator.PLUGIN_ID, "No preference key found for "+control)); //$NON-NLS-1$
      }
      if (enabledKeys != null && !enabledKeys.contains(prefName)) {
        value = false;
      }
    }
    enable(control, value);
  }

  /**
   * Enable the controls corresponding to the given set of keys.
   * For any key X, if there is a key X-check (the associated "Checkbox Key")
   * with a corresponding control, that control is also enabled.
   *
   * enabledKeys is updated with the set of all keys for all controls
   * that were enabled (including associated checkboxes).
   *
   * @param keys
   */
  private void setEnabledKeys(Collection<String> keys) {
    Set<String> enabledKeys = new HashSet<String>();
    if (keys != null) {
      for (String key : keys) {
        String enableCheckboxKeyFor = getEnableCheckboxKeyFor(key);
        if (fControls.containsKey(enableCheckboxKeyFor)) {
          enabledKeys.add(enableCheckboxKeyFor);
        }
      }
      enabledKeys.addAll(keys);
    }
    Set<Entry<String, Control>> entrySet = fControls.entrySet();
    for (Entry<String,Control> entry : entrySet) {
      boolean enable = enabledKeys.contains(entry.getKey());
      entry.getValue().setEnabled(enable);
//      entry.getValue().setVisible(enable);
    }
    this.enabledKeys = enabledKeys;
    updateSlaves();
  }

  @Override
  public void updateSlaves() {
    /* super.updateSlaves simply notifies all of the masters
     * of a selection event, so that all the slaves are
     * enabled/disable depending on the current master selection
     * status.
     */
    super.updateSlaves();
    for (Button[] exclusive : fExclusiveDependencies) {
      Button selected = getSelected(exclusive);
      if (selected == null) {
        selected = exclusive[0];
      }
      exclusiveSelectedEvent(exclusive, selected);
    }
  }

  public static Composite createScroll(Composite parent) {
    ScrolledPageContent scrolled= new ScrolledPageContent(parent);
    scrolled.setExpandHorizontal(true);
    scrolled.setExpandVertical(true);

    Composite rootComposite= scrolled.getBody();
    GridLayout layout= new GridLayout();
    layout.marginWidth= 0;
    layout.marginHeight= 0;
    rootComposite.setLayout(layout);
    scrolled.setContent(rootComposite);
    return rootComposite;
  }
}
