package com.objsys.asn1c.validator.preferences;

import static com.objsys.asn1c.validator.preferences.OptionPreference.createBooleanPreference;
import static com.objsys.asn1c.validator.preferences.OptionPreference.createComboPreference;
import static com.objsys.asn1c.validator.preferences.OptionPreference.createRadioButtonPreference;
import static com.objsys.asn1c.validator.preferences.OptionPreference.createPreference;
import static com.objsys.asn1c.validator.preferences.OptionPreference.createPreferenceWithCheck;
import static net.sourceforge.asneditor.ui.preferences.common.AbstractConfigurationBlock.getEnableCheckboxKeyFor;

import java.util.Arrays;
import java.util.Collection;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;

import net.sourceforge.asneditor.ui.preferences.common.OverlayPreferenceStore;
import net.sourceforge.asneditor.ui.preferences.common.OverlayPreferenceStore.OverlayKey;

import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.jface.preference.IPreferenceStore;

/**
 * ASN1CPreferenceConstants provides preference keys and is essentially a
 * factory for OptionPreference objects.
 * 
 * This class defines preference keys for all asn1c command line options.
 * 
 * This class also defines sets of keys (basic, Java, CSharp, etc).  These are 
 * used to help identify valid options (for target language etc.).
 * 
 * This class also defines a mapping from key to {@link OptionPreference}.  This 
 * mapping is initialized by a static initializer, and this is where you want
 * to create OptionPreference objects for new command line options.
 * 
 * Finally, this class has a number of private methods that create
 * dependencies between OptionPreferences.  These dependencies are used to
 * affect construction of the command line.
 * 
 * The dependencies work such that the primary preference is responsible for
 * generating the portion of the commmand line for itself and for its dependent
 * preferences, while the dependent preferences are configured as "internal" so
 * that they do not contribute directly to the command line.
 */
@SuppressWarnings("nls")
public abstract class ASN1CPreferenceConstants {
 public enum Language {
     
     LANG_JAVA("-java"), LANG_C("-c"), LANG_CPP("-c++"), LANG_C_SHARP("-c#"),
     LANG_VALIDATE("");
     
     private String value;  //value used for asn1c-language preference
     //except for LANG_VALIDATE, value is also the command line option
     
     
     private Language(String value ) { this.value = value; }
     public String toString() { return value; }
     
     /**
      * Return parameter associated with this language
      * @return
      */
     public String getParam() {
         if ( this == LANG_VALIDATE ) return "-syntaxcheck";
         else return value;
     }
     
     /**
      * Return Language associated with given preference value
      * @param value
      * @return
      */
     public static Language valueOfPref(String value) {
         Language[] values = Language.values();
         for (int i = 0; i < values.length; i++) {
            if ( values[i].value.equals(value) ) return values[i];
        }
        throw new IllegalArgumentException("No matching Language for '"
                + value + "'");
     } 
 };
 
// encoding rule options
  public static final String BER = "asn1c-ber";
  public static final String CER = "asn1c-cer";
  public static final String DER = "asn1c-der";
  public static final String APER = "asn1c-aper";
  public static final String UPER = "asn1c-uper";
  public static final String XER = "asn1c-xer";
  public static final String XML = "asn1c-xml";
  
  public static final String OER = "asn1c-oer";
  
  public static final String MDER = "asn1c-mder";
  
  public static final String JSON = "json";

  // language options
  public static final String LANGUAGE = "asn1c-language";
  public static final String LANG_VALIDATE = "";
  public static final String LANG_JAVA = "-java";
  public static final String LANG_C = "-c";
  public static final String LANG_CPP = "-c++";
  public static final String LANG_C_SHARP = "-c#";

  public static final String XSD = "asn1c-xsd";
  public static final String ASNSTD = "asn1c-asnstd";
  public static final String ASNSTD_MIXED = "mixed";
  public static final String ASNSTD_X208 = "x208";
  public static final String ASNSTD_X680 = "x680";

  // basic options
  public static final String USER_OPTIONS = "asn1c-user-options";
  public static final String COMPILER_PATH = "asn1c-path";
  public static final String COMPACT = "asn1c-compact";
  public static final String COMPAT = "asn1c-compat";
  public static final String COMPAT_CHECKED = "asn1c-compat-checked";
  public static final String CONFIG = "asn1c-config";
  public static final String DEPENDS = "asn1c-depends";
  public static final String GEN_ASN1 = "asn1c-gen-asn1";
  
  public static final String GEN_HTML = "asn1c-gen-html";
  
  public static final String IMPORT = "asn1c-I";
  public static final String LAX = "asn1c-lax";
  
  public static final String LAX_SYNTAX = "asn1c-lax-syntax";
  public static final String LIST = "asn1c-list";
  public static final String NODECODE = "asn1c-nodecode";
  public static final String NOENCODE = "asn1c-noencode";
  public static final String NO_INDEF_LEN = "asn1c-noindeflen";
  public static final String NO_OBJ_TYPES = "asn1c-noobjtypes";
  public static final String NO_OPEN_EXT = "asn1c-noopenext";
  public static final String NO_TYPES = "asn1c-notypes";
  public static final String OUTPUT_DIR = "asn1c-o";
  public static final String INPUT_FILES = "asn1c-input-files";
  public static final String PDU = "asn1c-pdu";
//  public static final String USE_PDU = "asn1c-usepdu";
  public static final String PRINT_NONE = "asn1c-print-none";
  public static final String PRINT = "asn1c-print";
  public static final String PRINT_CHECKED = "asn1c-print-checked";
  public static final String PRINT_FMT = "asn1c-printfmt";
  public static final String SHORT_NAMES = "asn1c-shortnames";
  public static final String TRACE = "asn1c-trace";
  public static final String NO_UNIQUE_NAMES = "asn1c-nouniquenames";
  public static final String WARNINGS = "asn1c-warnings";
  public static final String NO_DATE_STAMP = "asn1c-nodatestamp";

  public static final String NO_AMBIG_TAG = "asn1c-noambigtag";
  
  public static final String RAND_TEST = "asn1c-randtest";
  
  public static final String STRICT_SIZE = "asn1c-strictsize";

  // Java options
  public static final String JAVA_DIRS = "java-dirs";
  public static final String JAVA_GENBUILD = "java-genbuild";
  public static final String JAVA_GENANT = "java-genant";
  public static final String JAVA_GENJSOURCES = "java-genjsources";
  public static final String JAVA_GENMETA = "java-genmetadata";
  public static final String JAVA_GETSET = "java-getset";
  public static final String JAVA_PKGNAME = "java-pkgname";
  public static final String JAVA_PKGPFX = "java-pkgpfx";
  public static final String JAVA_SRC14 = "java-src14";
  public static final String JAVA_XMLPARSER = "java-xmlparser";
  
  public static final String JAVA_CLDC = "java-cldc";
  
  //The print option is materialized differently for java, where there is
  //just a simple checkbox.
  public static final String JAVA_PRINT = "java-print";
  public static final String JAVA_NOEVENTS = "java-noevents";

  // C/C++ options
  public static final String C_BINDIR = "c-bindir";
  public static final String C_HFILE = "c-hfile";
  public static final String C_CFILE = "c-cfile";
  public static final String C_CFILE_CHECKED = "c-cfile-checked";
  
  public static final String C_FULLY_QUAL_ENUMS = "c-fully-qual-enum";
  public static final String C_GEN_3G_L3 = "c-gen-3g-l3";
  
  public static final String C_GEN_BIT_MACROS = "c-genBitMacros";
  public static final String C_GENFREE = "c-genFree";
  public static final String C_HDR_GUARD = "c-hdr-guard";
  public static final String C_LIBDIR = "c-libdir";  
  public static final String C_MAXLINES = "c-maxlines";
  public static final String C_MAXLINES_CHECKED = "c-maxlines-checked";
  public static final String C_NAMESPACE = "c-namespace";
    
  public static final String C_NO_BITSTR32 = "c-nobitstr32";
  
  public static final String C_NO_ENUM_CONVERSION = "c-no_enum-conversion";
  
  public static final String C_NOINIT = "c-noInit";
  public static final String C_OBJDIR = "c-objdir";
  public static final String C_OH = "c-oh";
  public static final String C_STATIC = "c-static";
  
  public static final String C_TABLE_UNIONS = "c-table-unions";
  public static final String C_USE_ENUM_TYPES = "c-use-enum-types";
  
  public static final String C_CPP11 = "c-cpp11";
  
  public static final String C_PER_PAD_BITSTR = "c-perpad";
  public static final String C_PRJDIR = "c-prjdir";

  // C/C++ makefile/project options
  public static final String CM_BUILD_DLL = "cm-build-dll";
  
  public static final String CM_GENMAKE = "cm-genMake";
  public static final String CM_VCPROJ = "cm-vcproj";
  public static final String CM_VCPROJ_VER = "cm-vcproj-ver";
  
  public static final String CM_DLL = "cm-dll";
  public static final String CM_MT = "cm-mt";
  public static final String CM_W32 = "cm-w32";

  // C# options
  public static final String CSHARP_DIRS = "c#-dirs";
  public static final String CSHARP_NSPFX = "cs-nspfx";
  public static final String CSHARP_NAMESPACE = "cs-namespace";
  public static final String CSHARP_GENCS_SOURCES = "cs-gencs-sources";
  public static final String CSHARP_GENMETA = "cs-genmetadata";
  public static final String CSHARP_OUTPUT_SINGLE_FILE_NAME = "cs-output-single-file-name";
  
  public static final String CSHARP_OUTPUT_SINGLE_FILE = "cs-output-single-file";
  
  public static final String CSHARP_OUTPUT_BY_CLASS = "cs-output-by-class";
  
  public static final String CSHARP_OUTPUT_BY_MODULE = "cs-output-by-module";
  public static final String CSHARP_XMLPARSER = "cs-xmlparser";

  // Pro options
  public static final String PRO_3GPP = "pro-3gpp";
  
  public static final String PRO_EVENTS = "pro-events";
  public static final String PRO_STREAM = "pro-stream";
  public static final String PRO_TABLES = "pro-tables";
  public static final String PRO_STRICT = "pro-strict";
  public static final String PRO_PARAM = "pro-param";
  public static final String PRO_PRT2STR = "pro-prt2str";
  public static final String PRO_PRT2STRM = "pro-prt2strm";
  public static final String PRO_GENTEST = "pro-gentest";
  public static final String PRO_GENTEST_CHECKED = "pro-gentest-checked";
  public static final String PRO_COPY = "pro-copy";
  public static final String PRO_COPY_CHECKED = "pro-copy-checked";
  public static final String PRO_COMPARE = "java-compare";
  public static final String PRO_COMPARE_CHECKED = "java-compare-checked";

  public static final String PRO_READER = "pro-reader";
  public static final String PRO_WRITER = "pro-writer";
  public static final String PRO_MAXCFILES = "pro-maxcfiles";

  public static final String XSD_APPINFO = "xsd-appinfo";
  public static final String XSD_APPINFO_TAGS = "xsd-appinfo-tags";
  public static final String XSD_APPINFO_ENUMERATIONS = "xsd-appinfo-enums";
  public static final String XSD_APPINFO_EXTENSIONS = "xsd-appinfo-exts";

  public static final String XSD_ATTRS = "xsd-attrs";
  public static final String XSD_ATTRS_TAGS = "xsd-attrs-tags";
  public static final String XSD_ATTRS_ENUMERATIONS = "xsd-attrs-enums";
  public static final String XSD_ATTRS_EXTENSIONS = "xsd-attrs-exts";

  public static final String XSD_TARGETNS = "xsd-targetns";
  public static final String XSD_USE_ASN1XSD = "xsd-useAsn1Xsd";

  public static final String SYMBIAN = "symbian";
  public static final String SYMBIAN_DLL = "symbian-dll";

  public static final String[] SCHEMA_EXT = new String[]{"*.asn;*.asn1;*.txt", "*.xsd;*.wsdl", "*.*"};
  public static final String[] SCHEMA_EXT_NAMES = new String[] {
      "ASN.1 files (*.asn, *.asn1, *.txt)",
      "XSD and WSDL files(*.xsd, *.wsdl)", "All files" };

  /**
   * Each preference key has an associated OptionPreference which controls
   * how the value for the preference key is mapped to command line options.
   * key: preference key string
   * value: OptionPreference
   */
  private static final Map<String, OptionPreference> options = new Hashtable<String, OptionPreference>();

  // these keysets are used in order to validate/generate code for java/c/c++/c#/xsd
  /**
   * these preferences are used for all the generation kinds. (validate + generate)
   */
  private static String[] allwaysEnabledKeySet = new String[]{
    COMPILER_PATH, INPUT_FILES, ASNSTD, LANGUAGE, USER_OPTIONS, DEPENDS, WARNINGS, IMPORT, CONFIG, LIST,
    LAX_SYNTAX, GEN_ASN1, GEN_HTML, NO_AMBIG_TAG
  };

  /**
   * these preferences are used for all the generation kinds.
   */
  private static String[] basicKeySet = new String[]{
    // encoding rule options
    BER, CER, DER, APER, UPER, XER, XML, MDER, JSON,
    // basic options
    XSD, COMPACT, COMPAT, COMPAT_CHECKED, LAX,
    NODECODE, NOENCODE, NO_INDEF_LEN, NO_OBJ_TYPES, NO_OPEN_EXT, OUTPUT_DIR,
    PDU, PRINT_NONE, PRINT_CHECKED, PRO_COMPARE_CHECKED, SHORT_NAMES, 
    
    STRICT_SIZE, TRACE, NO_UNIQUE_NAMES, NO_DATE_STAMP,
    // Pro options
    PRO_EVENTS, PRO_STREAM, PRO_TABLES, PRO_STRICT, PRO_PARAM,
    PRO_READER, PRO_WRITER,
  };


  /**
   * these preferences are used for Validation
   */
  private static String[] validationKeySet = new String[]{

  };
  /**
   * these preferences are used for Java generation
   */
  private static String[] javaKeySet = new String[]{
    // Java options
    JAVA_DIRS, JAVA_GENBUILD, JAVA_GENANT, JAVA_GENJSOURCES, JAVA_GENMETA, 
    JAVA_GETSET, JAVA_PKGNAME, JAVA_PKGPFX, JAVA_SRC14,
    JAVA_CLDC, JAVA_NOEVENTS, JAVA_XMLPARSER,

    PRO_GENTEST_CHECKED
  };
  /**
   * these preferences are used for C/C++ generation
   */
  private static String[] cppKeySet = new String[]{
    // C/C++ options
    C_BINDIR, C_HFILE, C_CFILE, C_CFILE_CHECKED, C_FULLY_QUAL_ENUMS, 
    C_GEN_3G_L3, C_GEN_BIT_MACROS, C_GENFREE, C_HDR_GUARD, C_NAMESPACE, 
    C_LIBDIR,
    C_MAXLINES, C_MAXLINES_CHECKED, C_NO_BITSTR32, C_NO_ENUM_CONVERSION, 
    C_NOINIT, C_OBJDIR, C_OH, C_STATIC, C_TABLE_UNIONS, C_USE_ENUM_TYPES, OER, 
    RAND_TEST,
    C_CPP11, C_PER_PAD_BITSTR, C_PRJDIR,
    NO_TYPES,
    
    // C/C++ makefile/project options
    CM_BUILD_DLL, CM_GENMAKE, CM_VCPROJ, CM_VCPROJ_VER, CM_DLL, CM_MT, CM_W32,
    // pro
    PRINT, PRINT_FMT, PRO_3GPP, PRO_COMPARE, 
    PRO_PRT2STR, PRO_PRT2STRM, PRO_GENTEST_CHECKED, PRO_GENTEST, PRO_COPY, PRO_COPY_CHECKED, PRO_MAXCFILES,
    // Symbian options
    SYMBIAN
  };
  /**
   * these preferences are used for C# generation
   */
  private static String[] csharpKeySet = new String[]{
    // C# options
    CSHARP_NSPFX, CSHARP_NAMESPACE, CSHARP_GENCS_SOURCES, CM_GENMAKE, 
    CSHARP_GENMETA, CSHARP_DIRS, CSHARP_OUTPUT_BY_CLASS, 
    CSHARP_OUTPUT_BY_MODULE, CSHARP_OUTPUT_SINGLE_FILE, 
    CSHARP_OUTPUT_SINGLE_FILE_NAME, CSHARP_XMLPARSER
  };
  /**
   * these preferences are used for XSD generation
   */
  private static String[] xsdKeySet = new String[]{
    // XSD options
    XSD_APPINFO, XSD_APPINFO_TAGS, XSD_APPINFO_ENUMERATIONS, XSD_APPINFO_EXTENSIONS,
    XSD_ATTRS, XSD_ATTRS_TAGS, XSD_ATTRS_ENUMERATIONS, XSD_ATTRS_EXTENSIONS,
    XSD_TARGETNS, XSD_USE_ASN1XSD
  };
  /**
   * these preferences are used for Symbian generation
   */
  private static String[] symbianKeySet = new String[]{
    // Symbian options
    SYMBIAN_DLL
  };

  static {
    // encoding rule options
    addOption(createBooleanPreference(BER, false, Messages.ASN1CConfigurationBlock_BER));
    addOption(createBooleanPreference(CER, false, Messages.ASN1CConfigurationBlock_CER));
    addOption(createBooleanPreference(DER, false, Messages.ASN1CConfigurationBlock_DER));
    addOption(createBooleanPreference(APER, false, Messages.ASN1CConfigurationBlock_APER));
    addOption(createBooleanPreference(UPER, false, Messages.ASN1CConfigurationBlock_UPER));
    addOption(createBooleanPreference(XER, false, Messages.ASN1CConfigurationBlock_XER));
    addOption(createBooleanPreference(XML, false, Messages.ASN1CConfigurationBlock_XML));
    
    addOption(createBooleanPreference(OER, /*defaultValue=*/false, Messages.ASN1CConfigurationBlock_OER));
    
    addOption(createBooleanPreference(MDER, /*defaultValue=*/false, Messages.ASN1CConfigurationBlock_MDER));
    
    addOption(createBooleanPreference(JSON, /*defaultValue=*/false, Messages.ASN1CConfigurationBlock_JSON));

    // language options
    addOption(createComboPreference(LANGUAGE, new String[]{LANG_JAVA, LANG_C, LANG_CPP, LANG_C_SHARP, LANG_VALIDATE},
        new String[]{
        Messages.ASN1CConfigurationBlock_language_JAVA,
        Messages.ASN1CConfigurationBlock_language_C,
        Messages.ASN1CConfigurationBlock_language_CPP,
        Messages.ASN1CConfigurationBlock_language_CSHARP,
        Messages.ASN1CConfigurationBlock_language_NONE
        },
        LANG_JAVA, Messages.ASN1CConfigurationBlock_languages));

    // basic options
    addOption(createPreference(COMPILER_PATH, "", Messages.ASN1CConfigurationBlock_Compiler_path)); // parameter
    addOption(createPreference(USER_OPTIONS, "", Messages.ASN1CConfigurationBlock_User_options)); // parameter
    addOption(createComboPreference(ASNSTD, new String[]{ASNSTD_MIXED, ASNSTD_X208, ASNSTD_X680},
        new String[]{
        Messages.ASN1CConfigurationBlock_asnstd_mixed,
        Messages.ASN1CConfigurationBlock_asnstd_x208,
        Messages.ASN1CConfigurationBlock_asnstd_x680},
        ASNSTD_X680, Messages.ASN1CConfigurationBlock_AsnStd));
    addOption(createBooleanPreference(XSD, false, Messages.ASN1CConfigurationBlock_generate_XSD));
    
    addOption(createBooleanPreference(LAX_SYNTAX, false, Messages.ASN1ConfigurationBlock_lax_syntax));
    addOption(createBooleanPreference(COMPACT, false, Messages.ASN1CConfigurationBlock_Compact));
    addOption(createComboPreference(COMPAT, new String[]{"5.3", "5.4", "5.5", "5.6", "5.7", "5.8", "6.0",
          "6.1", "6.2", "6.3", "6.4"},
        new String[]{"v5.3x", "v5.4x", "v5.5x", "v5.6x", "v5.7x", "v5.8x", "v6.0x", "v6.1", "v6.2", "v6.3", "v6.4"},
        "6.4", 
        null));
    addOption(createBooleanPreference(COMPAT_CHECKED, false, Messages.ASN1CConfigurationBlock_Compat));
    addOption(createBooleanPreference(NO_AMBIG_TAG, false, Messages.ASN1CConfigurationBlock_noAmbigTag));

    addOption(createBooleanPreference(RAND_TEST, false, Messages.ASN1CConfigurationBlock_randTest));
    
    addOption(createBooleanPreference(STRICT_SIZE, false, Messages.ASN1CConfigurationBlock_strictSize));
    addOption(createPreference(CONFIG, "", Messages.ASN1CConfigurationBlock_Config));
    addOption(createBooleanPreference(DEPENDS, false, Messages.ASN1CConfigurationBlock_Depends));
    addOption( createPreferenceWithCheck( GEN_ASN1, "", Messages.ASN1CConfigurationBlock_gen_asn1));
    
    addOption( createBooleanPreference( GEN_HTML, /*defaultValue=*/false, Messages.ASN1CConfigurationBlock_gen_html));
    
    addOption(createPreference(IMPORT, "", Messages.ASN1CConfigurationBlock_Import));
    addOption(createBooleanPreference(LAX, false, Messages.ASN1CConfigurationBlock_Lax));
    addOption(createBooleanPreference(LIST, false, Messages.ASN1CConfigurationBlock_List));
    addOption(createBooleanPreference(NODECODE, true, Messages.ASN1CConfigurationBlock_NoDecode));
    addOption(createBooleanPreference(NOENCODE, true, Messages.ASN1CConfigurationBlock_NoEncode));
    addOption(createBooleanPreference(NO_INDEF_LEN, false, Messages.ASN1CConfigurationBlock_noIndefLen));
    addOption(createBooleanPreference(NO_OBJ_TYPES, false, Messages.ASN1CConfigurationBlock_noObjectTypes));
    addOption(createBooleanPreference(NO_OPEN_EXT, false, Messages.ASN1CConfigurationBlock_noOpenExt));
    addOption(createBooleanPreference(NO_TYPES, false, Messages.ASN1CConfigurationBlock_notypes));
    addOption(createPreference(OUTPUT_DIR, null, Messages.ASN1CConfigurationBlock_OutputDir));
    addOption(createPreference(INPUT_FILES, "", Messages.ASN1CConfigurationBlock_InputFiles));
    addOption(createPreference(PDU, "", Messages.ASN1CConfigurationBlock_pdu));
    // add an option writing to command line depending on PRINT
    //PRINT is the output file for the print methods (C/C++ only).
    addOption(createPreference(PRINT, "", null));
    addOption(createRadioButtonPreference(PRINT_NONE, true, Messages.ASN1CConfigurationBlock_noprint));

    //PRINT_CHECKED is the print to stdout(C/C++)/standard print (C#/Java) option
    addOption(createRadioButtonPreference(PRINT_CHECKED, false, 
          Messages.ASN1CConfigurationBlock_print));

    addOption(createRadioButtonPreference(PRO_PRT2STR, false, 
          Messages.ASN1CConfigurationBlock_pro_prtToStr));

    addOption(createRadioButtonPreference(PRO_PRT2STRM, false, 
          Messages.ASN1CConfigurationBlock_pro_prtToStrm));
    addOption(createComboPreference(PRINT_FMT, new String[]{"", "details", "bracetext"},
        new String[]{"default", "details", "bracetext"},
        "", Messages.ASN1CConfigurationBlock_prtfmt));
    
    addOption(createBooleanPreference(SHORT_NAMES, false, Messages.ASN1CConfigurationBlock_shortnames));
    addOption(createBooleanPreference(TRACE, false, Messages.ASN1CConfigurationBlock_trace));
    addOption(createBooleanPreference(NO_UNIQUE_NAMES, true, Messages.ASN1CConfigurationBlock_uniqueNames));
    addOption(createBooleanPreference(WARNINGS, false, Messages.ASN1CConfigurationBlock_warnings));
    addOption(createBooleanPreference(NO_DATE_STAMP, false, Messages.ASN1CConfigurationBlock_nodatestamp));

    // Java options
    addOption(createPreference(PRO_COMPARE, "", null));
    addOption(createBooleanPreference(PRO_COMPARE_CHECKED, false, Messages.ASN1CConfigurationBlock_pro_compare));

    addOption(createBooleanPreference(JAVA_DIRS, true, Messages.ASN1CConfigurationBlock_java_dirs));
    addOption(createBooleanPreference(JAVA_GENBUILD, false, Messages.ASN1CConfigurationBlock_java_genbuild));
    addOption(createBooleanPreference(JAVA_GENANT, false, Messages.ASN1CConfigurationBlock_java_genant));
    addOption(createBooleanPreference(JAVA_GENMETA, false, Messages.ASN1CConfigurationBlock_java_genmeta));
    addOption(createBooleanPreference(JAVA_GENJSOURCES, false, Messages.ASN1CConfigurationBlock_java_genjsources));
    addOption(createBooleanPreference(JAVA_GETSET, false, Messages.ASN1CConfigurationBlock_java_getset));
    addOption(createPreferenceWithCheck(JAVA_PKGPFX, "", Messages.ASN1CConfigurationBlock_java_pkgpfx));
    addOption(createPreferenceWithCheck(JAVA_PKGNAME, "", Messages.ASN1CConfigurationBlock_java_pkgname));
    addOption(createBooleanPreference(JAVA_SRC14, false, Messages.ASN1CConfigurationBlock_java_src14));
    
    addOption(createBooleanPreference(JAVA_CLDC, false, Messages.ASN1CConfigurationBlock_java_cldc));
    
    addOption(createBooleanPreference(JAVA_NOEVENTS, false, Messages.ASN1CConfigurationBlock_java_noevents));

    addOption( createComboPreference(JAVA_XMLPARSER, 
                   new String[] {"", "-stax", "-sax"},
                   new String[] {"XmlPull", "StAX", "SAX"},
                   "", Messages.ASN1CConfigurationBlock_java_xmlparser));
    
    // C/C++ options
    addOption( createBooleanPreference(C_FULLY_QUAL_ENUMS, false, Messages.ASN1CConfigurationBlock_c_fully_qual_enums) );
    
    addOption(createPreference(C_MAXLINES, "", Messages.ASN1CConfigurationBlock_c_maxlines));
    addOption(createRadioButtonPreference(C_MAXLINES_CHECKED, true, Messages.ASN1CConfigurationBlock_c_maxlines_checked));
    addOption(createRadioButtonPreference(PRO_MAXCFILES, false, Messages.ASN1CConfigurationBlock_pro_maxcfiles));
    addOption(createPreference(C_BINDIR, "", Messages.ASN1CConfigurationBlock_c_bindir));
    addOption(createPreference(C_HFILE, "", Messages.ASN1CConfigurationBlock_c_hfile));
    addOption(createPreference(C_CFILE, "", Messages.ASN1CConfigurationBlock_c_cfile));
    addOption(createRadioButtonPreference(C_CFILE_CHECKED, false, Messages.ASN1CConfigurationBlock_c_cfile_checked));

    addOption( createBooleanPreference(C_GEN_3G_L3, false, Messages.ASN1CConfigurationBlock_c_gen_3G_L3) );
    
    addOption(createBooleanPreference(C_GEN_BIT_MACROS, false, Messages.ASN1CConfigurationBlock_c_genBitMacros));
    addOption(createBooleanPreference(C_GENFREE, false, Messages.ASN1CConfigurationBlock_c_genFree));
    addOption( createPreferenceWithCheck(C_HDR_GUARD, "", Messages.ASN1CConfigurationBlock_c_hdr_guard) );
    addOption(createPreference(C_LIBDIR, "", Messages.ASN1CConfigurationBlock_c_libdir));    
    addOption( createPreferenceWithCheck(C_NAMESPACE, "", Messages.ASN1CConfigurationBlock_c_namespace) );
    
    addOption( createBooleanPreference(C_NO_BITSTR32, false, Messages.ASN1CConfigurationBlock_c_no_bitstr32) );
    
    addOption(createBooleanPreference(C_NO_ENUM_CONVERSION, false, Messages.ASN1CConfigurationBlock_c_no_enum_conversion));
    
    addOption(createBooleanPreference(C_NOINIT, false, Messages.ASN1CConfigurationBlock_c_noInit));
    addOption(createPreference(C_OBJDIR, "", Messages.ASN1CConfigurationBlock_c_objdir));    
    addOption(createPreference(C_OH, "", Messages.ASN1CConfigurationBlock_c_oh));
    addOption(createPreference(C_PRJDIR, "", Messages.ASN1CConfigurationBlock_c_prjdir));
    addOption(createBooleanPreference(C_STATIC, false, Messages.ASN1CConfigurationBlock_c_static));
    
    addOption(createBooleanPreference(C_TABLE_UNIONS, false, Messages.ASN1CConfigurationBlock_c_table_unions));

    addOption( createBooleanPreference(C_USE_ENUM_TYPES, false, Messages.ASN1CConfigurationBlock_c_use_enum_types) );
    
    addOption( createBooleanPreference(C_CPP11, false, Messages.ASN1CConfigurationBlock_c_cpp11) );
    
    addOption( createBooleanPreference(C_PER_PAD_BITSTR, false, Messages.ASN1CConfigurationBlock_c_per_pad_bitstr) );
    
    
    addOption(createBooleanPreference(PRO_3GPP, false, Messages.ASN1CConfigurationBlock_pro_3gpp) );

    addOption(createPreference(PRO_GENTEST, "", null));
    addOption(createBooleanPreference(PRO_GENTEST_CHECKED, false, Messages.ASN1CConfigurationBlock_pro_genTest));

    addOption(createPreference(PRO_COPY, "", null));
    addOption(createBooleanPreference(PRO_COPY_CHECKED, false, Messages.ASN1CConfigurationBlock_pro_copy));

    // C/C++ makefile/project options
    addOption(createPreference(CM_GENMAKE, "", Messages.ASN1CConfigurationBlock_cm_genMake));
    
    // If -vcproj is set, then we'll add the appropriate value to the command
    // line for the version.  Note that for VS 6.0, nothing extra is added
    // to the command line.
    addOption( createComboPreference(CM_VCPROJ_VER, 
          new String[] {"vc6", "2003", "2005", "2008", "2010", "2012", "", 
                         "2015"}, 
          new String[] {"VS 6.0", "VS 2003", "VS 2005", "VS 2008", "VS 2010", 
                         "VS 2012", "VS 2013", "VS 2015"},
          "", /* 2013 is current default */
          Messages.ASN1CConfigurationBlock_cm_vcproj_ver) );
    addOption(createBooleanPreference(CM_VCPROJ, false, Messages.ASN1CConfigurationBlock_cm_vcproj));
    addOption(createBooleanPreference(CM_BUILD_DLL, false, Messages.ASN1CConfigurationBlock_cm_build_dll));
    
    addOption(createBooleanPreference(CM_DLL, false, Messages.ASN1CConfigurationBlock_cm_dll));
    addOption(createBooleanPreference(CM_MT, false, Messages.ASN1CConfigurationBlock_cm_mt));
    addOption(createBooleanPreference(CM_W32, false, Messages.ASN1CConfigurationBlock_cm_w32));

    // C# options
    addOption(createBooleanPreference(CSHARP_DIRS, true, Messages.ASN1CConfigurationBlock_csharp_dirs));
    addOption(createPreferenceWithCheck(CSHARP_NSPFX, "", Messages.ASN1CConfigurationBlock_csharp_nspfx));
    addOption(createPreferenceWithCheck(CSHARP_NAMESPACE, "", Messages.ASN1CConfigurationBlock_csharp_namespace));
    addOption(createBooleanPreference(CSHARP_GENCS_SOURCES, false, Messages.ASN1CConfigurationBlock_csharp_gencssources));
    addOption(createBooleanPreference(CSHARP_GENMETA, false, Messages.ASN1CConfigurationBlock_csharp_genmeta));
    addOption( createRadioButtonPreference(CSHARP_OUTPUT_BY_CLASS, true, Messages.ASN1CConfigurationBlock_csharp_output_by_class));
    addOption( createRadioButtonPreference(CSHARP_OUTPUT_BY_MODULE, false, Messages.ASN1CConfigurationBlock_csharp_output_by_module));
    addOption( createRadioButtonPreference(CSHARP_OUTPUT_SINGLE_FILE, false, Messages.ASN1CConfigurationBlock_csharp_output_single_file));
    addOption(createPreference(CSHARP_OUTPUT_SINGLE_FILE_NAME, /*defaultValue=*/"", Messages.ASN1CConfigurationBlock_csharp_output_single_file_name));
    addOption( createComboPreference(CSHARP_XMLPARSER, 
          new String[] {"", "-sax"},
          new String[] {"XmlReader", "SAX"},
          "", Messages.ASN1CConfigurationBlock_csharp_xmlparser));

    // Pro options
    addOption(createBooleanPreference(PRO_EVENTS, false, Messages.ASN1CConfigurationBlock_pro_events));
    addOption(createBooleanPreference(PRO_STREAM, false, Messages.ASN1CConfigurationBlock_pro_stream));
    addOption(createBooleanPreference(PRO_TABLES, false, Messages.ASN1CConfigurationBlock_pro_tables));
    addOption(createBooleanPreference(PRO_STRICT, false, Messages.ASN1CConfigurationBlock_pro_strict));
    addOption(createPreference(PRO_PARAM, "", Messages.ASN1CConfigurationBlock_pro_param));

    addOption(createBooleanPreference(PRO_READER, false, Messages.ASN1CConfigurationBlock_pro_reader));
    addOption(createBooleanPreference(PRO_WRITER, false, Messages.ASN1CConfigurationBlock_pro_writer));

    // XSD options
    addOption(createBooleanPreference(XSD_APPINFO_TAGS, true, Messages.ASN1CConfigurationBlock_xsd_appinfo_tags));
    addOption(createBooleanPreference(XSD_APPINFO_ENUMERATIONS, true, Messages.ASN1CConfigurationBlock_xsd_appinfo_enumerations));
    addOption(createBooleanPreference(XSD_APPINFO_EXTENSIONS, true, Messages.ASN1CConfigurationBlock_xsd_appinfo_extensions));
    addOption(createBooleanPreference(XSD_APPINFO, false, Messages.ASN1CConfigurationBlock_xsd_appinfo));

    addOption(createBooleanPreference(XSD_ATTRS_TAGS, true, Messages.ASN1CConfigurationBlock_xsd_attrs_tags));
    addOption(createBooleanPreference(XSD_ATTRS_ENUMERATIONS, true, Messages.ASN1CConfigurationBlock_xsd_attrs_enumerations));
    addOption(createBooleanPreference(XSD_ATTRS_EXTENSIONS, true, Messages.ASN1CConfigurationBlock_xsd_attrs_extensions));
    addOption(createBooleanPreference(XSD_ATTRS, false, Messages.ASN1CConfigurationBlock_xsd_attrs));

    addOption(createPreference(XSD_TARGETNS, "", Messages.ASN1CConfigurationBlock_xsd_targetns));
    addOption(createPreference(XSD_USE_ASN1XSD, "", Messages.ASN1CConfigurationBlock_xsd_use_asn1xsd));

    // Symbian options
    addOption(createBooleanPreference(SYMBIAN, false, Messages.ASN1CConfigurationBlock_symbian_opt));
    addOption(createBooleanPreference(SYMBIAN_DLL, false, Messages.ASN1CConfigurationBlock_symbian_opt_dll));
  }

  public static final OptionPreference getCommandLineOption(String key) {
    return options.get(key);
  }

  public static boolean isCOrCppLanguage(IPreferenceStore store) {
    String language = store.getString(ASN1CPreferenceConstants.LANGUAGE);
    return (LANG_C.equals(language) || LANG_CPP.equals(language));
  }

  public static boolean isJavaOrCSharpLanguage(IPreferenceStore store) {
    String language = store.getString(ASN1CPreferenceConstants.LANGUAGE);
    return (LANG_JAVA.equals(language) || LANG_C_SHARP.equals(language));
  }
//
//  /**
//   * Create an Option whose command line, if the primary preference is set,
//   * consists of the primary option, followed by a comma-separated list
//   * of the options for the secondary options.
// * @param preferenceKey
// * @param defaultValue
// * @param label
//   * 
//   * @return
//   */
//  private static OptionPreference createBooleanOptionPreference(
//
//      String preferenceKey, boolean defaultValue,
//
//      String label) {
//
//    OptionPreference combinedOptionPreference = 
//       new BooleanOptionPreference(preferenceKey, defaultValue, 
//             label);
//
//      return combinedOptionPreference;
//
//  }

  /*
   * Add an OptionPreference to the options map.
   * May add an additional option for an enabling check box.
   */
  private static void addOption(OptionPreference optionPreference) {
    options.put(optionPreference.getPreferenceKey(), optionPreference);
    if (optionPreference.isAddEnableCheckBox()) {
      options.put(getEnableCheckboxKeyFor(optionPreference.getPreferenceKey()), optionPreference);
    }
  }

  public static final Collection<String> getPreferenceKeys() {
    return new LinkedList<String>(options.keySet());
  }

  public static Collection<String> getPreferenceKeys(String language, boolean enableXSD, boolean enableSymbian) {
    if (LANG_C.equals(language)) { //$NON-NLS-1$
      return getCPreferenceKeys(enableXSD, enableSymbian);
    } else if (LANG_CPP.equals(language)) { //$NON-NLS-1$
      return getCppPreferenceKeys(enableXSD, enableSymbian);
    } else if (LANG_C_SHARP.equals(language)) { //$NON-NLS-1$
      return getCSharpPreferenceKeys(enableXSD);
    } else if (LANG_JAVA.equals(language)) { //$NON-NLS-1$
      return getJavaPreferenceKeys(enableXSD);
    } else { // validation
      return getValidationPreferenceKeys();
    }
  }

  public static final Collection<String> getValidationPreferenceKeys() {
    LinkedList<String[]> res = new LinkedList<String[]>();
    res.add(allwaysEnabledKeySet);
    res.add(validationKeySet);
    Collection<String> result = makeCollectionWithoutDupplicates(res);
    result.remove(LANGUAGE); // no language needed for validation
    return result;
  }

  public static final Collection<String> getJavaPreferenceKeys(boolean enableXSD) {
    LinkedList<String[]> res = new LinkedList<String[]>();
    res.add(allwaysEnabledKeySet);
    res.add(basicKeySet);
    res.add(javaKeySet);
    if (enableXSD) {
      res.add(xsdKeySet);
    }
    return makeCollectionWithoutDupplicates(res);
  }

  public static final Collection<String> getCPreferenceKeys(boolean enableXSD, boolean enableSymbian) {
    return getCppPreferenceKeys(enableXSD, enableSymbian);
  }

  public static final Collection<String> getCppPreferenceKeys(boolean enableXSD, boolean enableSymbian) {
    LinkedList<String[]> res = new LinkedList<String[]>();
    res.add(allwaysEnabledKeySet);
    res.add(basicKeySet);
    res.add(cppKeySet);
    if (enableXSD) {
      res.add(xsdKeySet);
    }
    if (enableSymbian) {
      res.add(symbianKeySet);
    }
    return makeCollectionWithoutDupplicates(res);
  }

  public static final Collection<String> getCSharpPreferenceKeys(boolean enableXSD) {
    LinkedList<String[]> res = new LinkedList<String[]>();
    res.add(allwaysEnabledKeySet);
    res.add(basicKeySet);
    res.add(csharpKeySet);
    if (enableXSD) {
      res.add(xsdKeySet);
    }
    return makeCollectionWithoutDupplicates(res);
  }

  public static Collection<String> getAlwaysEnabledKeys() {
    LinkedList<String[]> res = new LinkedList<String[]>();
    res.add(allwaysEnabledKeySet);
    return makeCollectionWithoutDupplicates(res);
  }

  private static Collection<String> makeCollectionWithoutDupplicates(Collection<String[]> keySets) {
    String[][] keys = keySets.toArray(new String[0][]);
    return makeCollectionWithoutDupplicates(keys);
  }

  private static Collection<String> makeCollectionWithoutDupplicates(String[][] keySets) {
    Collection<String> res = new LinkedList<String>();
    for (String[] set : keySets) {
      res.addAll(Arrays.asList(set));
    }
    return res;
  }

  /**
   * Put default values for all OptionPreferences into given preferences object.
   * This method is used by {@link PreferenceInitializer} to initialize the
   * plug-in's default preference store. 
   * @param prefs
   */
  public static final void loadDefaults(IEclipsePreferences prefs) {
     
     Collection<String> preferenceKeys = 
        ASN1CPreferenceConstants.getPreferenceKeys();

     for (String key : preferenceKeys) {

       OptionPreference option = 
          ASN1CPreferenceConstants.getCommandLineOption(key);

       Object defaultValue = option.getDefaultValue();
       prefs.put(key, defaultValue == null ? "" : defaultValue.toString());
     }
  }
     
  
  /**
   * Create an array of OverlayKey objects, with an OverlayKey for each
   * preference key and default value.
   * 
   * The OverlayPreferenceStore class holds an array of overlay keys, which
   * are used by its loadDefaults method.  That method is invoked when the
   * preferences in the store need to be reset to default values.  It is NOT
   * used to provide default values initially.  For initial default values,
   * see {@link PreferenceInitializer}. 
   * 
   * @return
   */
  public static final OverlayPreferenceStore.OverlayKey[] createOverlayStoreKeys() {
    Collection<String> preferenceKeys = ASN1CPreferenceConstants.getPreferenceKeys();

    LinkedList<OverlayPreferenceStore.OverlayKey> res = new LinkedList<OverlayPreferenceStore.OverlayKey>();

    for (String key : preferenceKeys) {
      OptionPreference option = ASN1CPreferenceConstants.getCommandLineOption(key);
      final OverlayKey overlayKey;
      if (option.isBooleanOption()) {
        overlayKey = OverlayPreferenceStore.createBoolean(key);
      } else {
        overlayKey = OverlayPreferenceStore.createString(key);
      }
      overlayKey.setDefaultValue(option.getDefaultValue());
      res.add(overlayKey);
    }

    return res.toArray(new OverlayPreferenceStore.OverlayKey[0]);
  }
}
