package com.objsys.asn1c.validator.preferences;

import org.eclipse.core.runtime.Assert;

/**
 * An OptionPreference provides information on how to present a preference
 * to the user (e.g., as a checkbox, with what label).
 *  
 * A preference may have an associated enable check box.  This is useful
 * for options that have optional values.  For example, the -compare <filename>
 * option is presented as a checkbox combined with a text field.  In this case
 * two preferences are used: one for the check box and one for the value.  Thus,
 * the check box can be cleared and the string value can remain but not be
 * effective (ie not be used on the command line).
 *  
 * Numerous subclasses provide specializations (e.g. BooleanPreference).
 * 
 * A number of static convenience methods exist (e.g. createBooleanPreference).
 *
 */
public class OptionPreference {

  protected String label;
  protected final String preferenceKey;

  private Object defaultValue;
  private boolean addEnableCheckBox;
  private boolean radioButton;

  /**
   * Represents a boolean-valued preference.
   * 
   * When the preference is set true, the command line option will be included
   * on the command line.  If {@code inverse} is true, then the option is
   * included only if the preference is false.
   */
  public static class BooleanOptionPreference extends OptionPreference {
    BooleanOptionPreference(String preferenceKey, boolean defaultValue, String label) {
      super(preferenceKey, defaultValue, label);
    }

    public boolean isBooleanOption() {
      return true;
    }
  }


  /**
   * Adds a list of values and corresponding labels to a StringOptionPreference.
   */
  public static class ComboOptionPreference extends OptionPreference {
    String[] values;
    String[] valueLabels;

    ComboOptionPreference(String preferenceKey, String[] values, String[] valueLabels, String defaultValue, String label) {
      super(preferenceKey, defaultValue, label);
      this.values = values;
      this.valueLabels = valueLabels;
      Assert.isLegal(values != null, "Invalid values NULL."); //$NON-NLS-1$
      Assert.isLegal(valueLabels != null, "Invalid value labels NULL."); //$NON-NLS-1$
      Assert.isLegal(values.length == valueLabels.length, "Values and value labels must have the same length."); //$NON-NLS-1$
    }

    public String[] getValues() {
      return values;
    }

    public String[] getValueLabels() {
      return valueLabels;
    }

    public boolean isComboOption() {
      return values != null;
    }
  }

  public OptionPreference(String preferenceKey, Object defaultValue, String label) {
    this.preferenceKey = preferenceKey;
    this.defaultValue = defaultValue;
    this.label = label;
    this.addEnableCheckBox = false;
  }

  /**
   * @return a boolean option instance generating the command line option when the value is true (e.g. -genant)
   */
  public static BooleanOptionPreference createBooleanPreference(String preferenceKey, boolean defaultValue, String label) {
    return new BooleanOptionPreference(preferenceKey, defaultValue, label);
  }

  /**
   * @return a boolean option instance generating the command line option when the value is true (e.g. -genant)
   */
  public static BooleanOptionPreference createRadioButtonPreference(String preferenceKey, boolean defaultValue, String label) {
    BooleanOptionPreference booleanOptionPreference = new BooleanOptionPreference(preferenceKey, defaultValue, label);
    booleanOptionPreference.setRadioButton(true);
    return booleanOptionPreference;
  }


  public static OptionPreference createComboPreference(String preferenceKey, String[] values, String[] valueLabels, String defaultValue, String label) {
    return new ComboOptionPreference(preferenceKey, values, valueLabels, defaultValue, label);
  }

  public static OptionPreference createPreference(String preferenceKey, String defaultValue, String label) {
    return new OptionPreference(preferenceKey, defaultValue, label);
  }
  /**
   * Create an OptionPreference for a preference that has a related
   * "check" preference.
   * @param preferenceKey
   * @param defaultValue
   * @param label
   * @return OptionPreference  
   */
  public static OptionPreference createPreferenceWithCheck(String preferenceKey, String defaultValue, String label) {
    OptionPreference option = createPreference(preferenceKey, defaultValue, label);
    option.addEnableCheckBox = true;
    return option;
  }

  public String getLabel() {
    return label;
  }

  
  public String getPreferenceKey() {
    return preferenceKey;
  }

  public boolean isBooleanOption() {
    return false;
  }


  public boolean isComboOption() {
    return false;
  }

  public Object getDefaultValue() {
    return defaultValue;
  }

  public boolean isAddEnableCheckBox() {
    return addEnableCheckBox;
  }

  public String toString() {
    return getPreferenceKey()+" "+getLabel();  //$NON-NLS-1$//$NON-NLS-2$
  }

  public boolean isRadioButton() {
    return radioButton;
  }

  public void setRadioButton(boolean radioButton) {
    this.radioButton = radioButton;
  }

  public void setLabel(String label) {this.label = label;}
}
