package com.objsys.asn1c.validator.project;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import net.sourceforge.asneditor.ui.preferences.common.OverlayPreferenceStore;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jface.preference.IPreferenceStore;

import com.objsys.asn1c.Activator;
import com.objsys.asn1c.validator.lib.ASN1CClasspathContainerInitializer;
import com.objsys.asn1c.validator.preferences.ASN1CPreferenceConstants;
import com.objsys.asn1c.validator.wizard.FileUtils;


public class ACProject {
  public static final String ASN1C_PROJECT_PROPERTIES_FILE = ".asn1c.project"; //$NON-NLS-1$
  public static final String DEFAULT_SCHEMA_PATH = "asn1"; //$NON-NLS-1$

  private IProject project;
  private OverlayPreferenceStore preferenceStore;
  private String filename = ASN1C_PROJECT_PROPERTIES_FILE;

  private String schemaPath = DEFAULT_SCHEMA_PATH;

  public static final String PATH_SEPARATOR = File.pathSeparator;

  public ACProject(IProject project) {
    setProject(project);
    preferenceStore = new OverlayPreferenceStore(
        Activator.getDefault().getPreferenceStore(),
        ASN1CPreferenceConstants.createOverlayStoreKeys()){

          @Override
          public void addKeys(OverlayKey[] keys) {
            // ignore updates, since already initialized
          }};

    preferenceStore.load();
    preferenceStore.start();
  }

  public IProject getProject() {
    return project;
  }

  public void setProject(IProject project) {
    this.project = project;
  }

  public String getSchemaPath() {
    return schemaPath;
  }

  public void setSchemaPath(String schemaPath) {
    this.schemaPath = schemaPath;
  }

  public OverlayPreferenceStore getPreferenceStore() {
    return preferenceStore;
  }

  public void loadPreferenceStore() throws CoreException, IOException {
    IFile file = project.getFile(new Path(filename));
    InputStream in = file.getContents();
    try {
      Properties properties = new Properties();
      properties.load(in);
      preferenceStore.setFromProperties(properties);
      preferenceStore.setFromParent(ASN1CPreferenceConstants.COMPILER_PATH);
    } finally {
      try {
        in.close();
      } catch (Exception e) {
        // ignore
      }
    }
  }

  public void save() throws CoreException, IOException {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    Properties properties = preferenceStore.getAsProperties();
    properties.store(out, "ASN1CProject created "+new Date()); //$NON-NLS-1$
    out.flush();
    out.close();
    InputStream source = new ByteArrayInputStream(out.toByteArray());
    createFile(source, filename);
  }

  public IFile copyFile(String systemPath, String projectPath) throws CoreException, IOException {
    FileInputStream source = new FileInputStream(systemPath);
    return createFile(source, projectPath);
  }

  private IFile createFile(InputStream source, String projectPath)
      throws CoreException, FileNotFoundException {
    IFile file = project.getFile(new Path(projectPath));
    if (file.exists()) {
      file.delete(true, new NullProgressMonitor());
    }
    try {
      file.create(source, true, new NullProgressMonitor());
    } finally {
      try {
        source.close();
      } catch (Exception e) {
        // ignore
      }
    }
    return file;
  }
  public void addToCollection(String key, String delim, String input) {
    addToCollection(preferenceStore, key, delim, input);
  }
  public static void addToCollection(IPreferenceStore store, String key, String delim, String input) {
    if (input == null)
      return;
    Collection<String> collection = getCollection(store, key, delim);
    collection.add(input);
    setCollection(store, key, delim, collection);
  }

  public void setCollection(String key, String delim, Collection<String> collection) {
    setCollection(preferenceStore, key, delim, collection);
  }

  public static void setCollection(IPreferenceStore store, String key, String delim, Collection<String> collection) {
    if (collection == null) {
      store.setValue(key, ""); //$NON-NLS-1$
      return;
    }
    StringBuilder sb = new StringBuilder();
    int size = collection.size();
    for (String string : collection) {
      sb.append(string);
      --size;
      if (size > 0)
        sb.append(delim);
    }
    store.setValue(key, sb.toString());
  }

  public Collection<String> getCollection(String key, String delim) {
    return getCollection(preferenceStore, key, delim);
  }

  public static Collection<String> getCollection(IPreferenceStore store, String key, String delim) {
    String string = store.getString(key);
    return toArray(string, delim);
  }


  public static Collection<String> toArray(String string, String delim) {
    HashSet<String> res = new HashSet<String>();
    if (string != null) {
      StringTokenizer tk = new StringTokenizer(string, delim, false);
      while (tk.hasMoreTokens()) {
        String file = tk.nextToken();
        if (file.trim().length() > 0)
          res.add(file);
      }
    }
    return res;
  }

  /**
   * Creates a project resource handle for the current project name field
   * value.
   * <p>
   * This method does not create the project resource; this is the
   * responsibility of <code>IProject::create</code> invoked by the new
   * project resource wizard.
   * </p>
   * @param projectName
   *
   * @return the new project resource handle
   */
  public static IProject getProjectHandle(String projectName) {
    return ResourcesPlugin.getWorkspace().getRoot().getProject(projectName);
  }

  public static ACProject create(IProject projectHandle) throws CoreException {
    @SuppressWarnings("unused")
    IJavaProject javaPrj = JavaCore.create(projectHandle);
    ASN1CNature.addNature(projectHandle);
    ACProject acPrj = new ACProject(projectHandle);
    return acPrj;
  }

  public IFolder createFolder(String path) throws CoreException {
    IFolder folder = project.getFolder(new Path(path));
    if (!folder.exists()) {
      folder.create(true, true, new NullProgressMonitor());
    }
    return folder;
  }

  public void copyFilesToWorkspace() throws CoreException, IOException {
    copySchemaFilesToWorkspace();

    String configFile = getPreferenceStore().getString(ASN1CPreferenceConstants.CONFIG);
    if (configFile != null && configFile.trim().length() > 0)
      copyToWorkspace(configFile, "/"); //$NON-NLS-1$
  }

  public void copySchemaFilesToWorkspace() throws CoreException, IOException {
    createFolder(getSchemaPath());
    Collection<String> files = getCollection(ASN1CPreferenceConstants.INPUT_FILES, ACProject.PATH_SEPARATOR);
    HashSet<String> res = new HashSet<String>();
    for (String filename : files) {
      filename = copyToWorkspace(filename, getSchemaPath());
      res.add(filename);
    }
    setCollection(ASN1CPreferenceConstants.INPUT_FILES, ACProject.PATH_SEPARATOR, res);
  }

  public String copyToWorkspace(String filename, String destPath) throws CoreException, IOException {
    IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
    if (!root.exists(new Path(filename)) && !FileUtils.isInWorkspace(filename)) {
      String resourceName = destPath+"/"+new Path(filename).lastSegment(); //$NON-NLS-1$
      IFile resultFile = copyFile(new File(filename).getAbsolutePath(), resourceName);
      filename = resultFile.getFullPath().toPortableString();
    }
    return filename;
  }

  public static void addJarsToClassPath(IJavaProject javaProject) throws CoreException {

    IClasspathEntry[] entries = javaProject.getRawClasspath();
    for (IClasspathEntry classpathEntry : entries) {
      if (classpathEntry.getEntryKind() == IClasspathEntry.CPE_CONTAINER &&
          ASN1CClasspathContainerInitializer.ASN1C_LIBS_PATH.equals(classpathEntry.getPath())) {
        // nothing to do, library already exists
        return;
      }
    }
    List<IClasspathEntry> cp= new ArrayList<IClasspathEntry>(entries.length + 1);
    cp.addAll(Arrays.asList(entries));

    IClasspathEntry[] defaultEntries = getDefaultClasspathEntries();
    cp.addAll(Arrays.asList(defaultEntries));

    javaProject.setRawClasspath(cp.toArray(new IClasspathEntry[0]), new NullProgressMonitor());
  }

  public static IClasspathEntry[] getDefaultClasspathEntries() {
    List<IClasspathEntry> cp= new ArrayList<IClasspathEntry>(1);
    cp.add(JavaCore.newContainerEntry(ASN1CClasspathContainerInitializer.ASN1C_LIBS_PATH));
    return cp.toArray(new IClasspathEntry[0]);
  }
}
